# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Utils for Radial Anaylsis."""

import math
import numpy as np
import numpy.linalg as la
from tqdm import tqdm
from hyperspy.components1d import Polynomial, Gaussian
from hyperspy.signals import Signal2D
from hyperspy.misc.utils import isiterable
import hyperspy.api as hs
import pyxem.utils._pixelated_stem_tools as pst
from pyxem.utils._deprecated import deprecated


@deprecated(since="0.18.0", removal="0.19.0")
def _centre_comparison(s, steps, step_size, crop_radial_signal=None, angleN=8):
    """
    Compare how the centre position affects the radial average.
    Useful for finding on optimal centre position if one has some
    round feature. Returns a list of image stacks, one image stack
    for each centre position, and the image stack consists of radially
    averaged segments of an image.

    Currently only works with signals with 0 navigation dimensions.

    Parameters
    ----------
    s : HyperSpy 2D signal
    steps : int
    step_size : int
    crop_radial_signal : tuple, optional
    angleN : int, default 8
    """
    if s.axes_manager.navigation_dimension != 0:
        raise ValueError(
            "centre_comparison only works for pixelatedSTEM "
            "signals with 0 navigation dimensions"
        )
    s_list = []
    centre_list = get_centre_position_list(s, steps, step_size)
    for centre_x, centre_y in centre_list:
        s_angle = s.angular_slice_radial_average(
            angleN=angleN,
            centre_x=np.array([centre_x]),
            centre_y=np.array([centre_y]),
            show_progressbar=False,
        )
        if crop_radial_signal is not None:
            s_angle = s_angle.isig[crop_radial_signal[0] : crop_radial_signal[1]]
        s_angle.metadata.add_node("Angle_slice_processing")
        s_angle.metadata.Angle_slice_processing.centre_x = centre_x
        s_angle.metadata.Angle_slice_processing.centre_y = centre_y
        s_list.append(s_angle)
    return s_list


@deprecated(since="0.18.0", removal="0.19.0")
def get_coordinate_of_min(s):
    """Returns the x and y values of the minimum in a signal."""
    z = s.data
    idx = np.argwhere(z == np.min(z))
    x = s.axes_manager[0].index2value(idx[0][1])
    y = s.axes_manager[1].index2value(idx[0][0])
    return (x, y)


@deprecated(since="0.18.0", removal="0.19.0")
def get_centre_position_list(s, steps, step_size):
    """
    Returns a zip of x and y coordinates based on the offset of the center-
    point, and number of steps in each direction and a step-size given in
    pixels. Scale and offset taken from axes_manager.
    """
    scale_x = s.axes_manager.signal_axes[0].scale
    scale_y = s.axes_manager.signal_axes[1].scale
    d1_x = scale_x * steps * step_size
    d2_x = scale_x * (steps + 1) * step_size
    d1_y = scale_y * steps * step_size
    d2_y = scale_y * (steps + 1) * step_size

    x0 = -s.axes_manager[0].offset
    y0 = -s.axes_manager[1].offset
    centre_x_list, centre_y_list = [], []
    range_x = np.arange(x0 - d1_x, x0 + d2_x, step_size * scale_x)
    range_y = np.arange(y0 - d1_y, y0 + d2_y, step_size * scale_y)
    for x in range_x:
        for y in range_y:
            centre_x_list.append(x)
            centre_y_list.append(y)
    centre_list = zip(centre_x_list, centre_y_list)
    return centre_list


@deprecated(since="0.18.0", removal="0.19.0")
def get_optimal_centre_position(
    s, radial_signal_span, steps=3, step_size=1, angleN=8, show_progressbar=True
):
    """
    Find centre position of a ring by using angle sliced radial average.

    Takes signal s, radial span of feature used to determine the centre
    position. Radially averages the feature in angleN segments, for
    each possible centre position. Models this averaged signal
    as a Gaussian and returns array of with the standard deviations
    of these Gaussians.

    Note, the approximate centre position must be set using the offset
    parameter in the signal. The offset parameter is the negative of the
    position shown with s.plot(). So if the centre position in the image is
    x=52 and y=55:
    s.axes_manager.signal_axes[0].offset = -52
    s.axes_manager.signal_axes[1].offset = -55
    This can be checked by replotting the signal, to see if the centre
    position has the values x=0 and y=0.

    Parameters
    ----------
    s : HyperSpy 2D signal
        Only supports signals with no navigation dimensions.
    radial_signal_span : tuple
        Range for finding the circular feature.
    steps : int, default 3
        Number of steps in x/y direction to look for the optimal
        centre position. If the offset is (50, 55), and step_size 1:
        the positions x=(45, 46, 47, 48, ..., 55) and y=(50, 51, 52, ..., 60),
        will be checked.
    step_size : int, default 1
    angleN : int, default 8
        See angular_slice_radial_average for information about angleN.

    Returns
    -------
    Optimal centre position : HyperSpy 2D signal

    Examples
    --------
    >>> s = pxm.data.dummy_data.get_single_ring_diffraction_signal()
    >>> s.axes_manager.signal_axes[0].offset = -105
    >>> s.axes_manager.signal_axes[1].offset = -67
    >>> import pyxem.utils.radial_utils as ra
    >>> s_ocp = ra.get_optimal_centre_position(s, (35, 45), steps=2)
    >>> centre_pos = ra.get_coordinate_of_min(s_ocp)

    """
    s_list = _centre_comparison(
        s, steps, step_size, crop_radial_signal=radial_signal_span, angleN=angleN
    )

    m_list = []
    for temp_s in tqdm(s_list, disable=(not show_progressbar)):
        temp_s.change_dtype("float64")
        temp_s = temp_s - temp_s.data.min()
        temp_s = temp_s / temp_s.data.max()
        am = temp_s.axes_manager
        g_centre = temp_s.inav[0].data.argmax() * am[-1].scale + am[-1].offset
        g_sigma = 3
        g_A = temp_s.data.max() * 2 * g_sigma

        m = temp_s.create_model()
        g = Gaussian(A=g_A, centre=g_centre, sigma=g_sigma)
        m.append(g)
        m.multifit(show_progressbar=False)
        m_list.append(m)
    s_centre_std_array = _get_offset_image(m_list, s, steps, step_size)
    return s_centre_std_array


@deprecated(since="0.18.0", removal="0.19.0")
def refine_signal_centre_position(
    s, radial_signal_span, refine_step_size=None, **kwargs
):
    """Refine centre position of a diffraction signal by using round feature.

    This function simply calls get_optimal_centre_position with a
    smaller and smaller step_size, giving a more accurate centre position.

    The calculated centre position is stored in the offset value in the
    input signal.

    Parameters
    ----------
    s : HyperSpy 2D signal
        Approximate centre position must be set beforehand, see
        get_optimal_centre_position for more information.
    radial_signal_span : tuple
    refine_step_size : list, default [1, 0.5, 0.25]
    **kwargs
        Passed to get_optimal_centre_position

    See Also
    --------
    get_optimal_centre_position : finds the centre positions for a step_size

    Example
    -------
    >>> import pyxem.utils.radial_utils as ra
    >>> s = pxm.data.dummy_data.get_single_ring_diffraction_signal()
    >>> s.axes_manager[0].offset, s.axes_manager[1].offset = -103, -69
    >>> ra.refine_signal_centre_position(s, (32., 48.), angleN=4)
    >>> x, y = s.axes_manager[0].offset, s.axes_manager[1].offset

    """
    if refine_step_size is None:
        refine_step_size = [1.0, 0.5, 0.25]
    for step_size in refine_step_size:
        s_centre = get_optimal_centre_position(
            s=s, radial_signal_span=radial_signal_span, step_size=step_size, **kwargs
        )
        x, y = get_coordinate_of_min(s_centre)
        s.axes_manager[0].offset, s.axes_manager[1].offset = -x, -y


@deprecated(since="0.18.0", removal="0.19.0")
def _get_offset_image(model_list, s, steps, step_size):
    """
    Make offset image signal based on difference in Gaussian centre position.

    Creates a Signal2D object of the standard deviation of the
    Gaussians fitted to the radially averaged features, as a
    function of center coordinate.

    Parameters
    ----------
    model_list : list of HyperSpy model objects
    s : HyperSpy 2D signal
    steps : int
    step_size : int

    Returns
    -------
    offset_signal : HyperSpy 2D signal

    """
    s_offset = Signal2D(np.zeros(shape=((steps * 2) + 1, (steps * 2) + 1)))
    am = s.axes_manager.signal_axes

    offset_x0 = -am[0].offset - (steps * step_size)
    offset_y0 = -am[1].offset - (steps * step_size)
    s_offset.axes_manager.signal_axes[0].offset = offset_x0
    s_offset.axes_manager.signal_axes[1].offset = offset_y0
    s_offset.axes_manager.signal_axes[0].scale = step_size
    s_offset.axes_manager.signal_axes[1].scale = step_size
    am_o = s_offset.axes_manager.signal_axes
    for model in model_list:
        x = model.signal.metadata.Angle_slice_processing.centre_x
        y = model.signal.metadata.Angle_slice_processing.centre_y
        std = model.components.Gaussian.centre.as_signal().data.std()
        iX, iY = am_o[0].value2index(x), am_o[1].value2index(y)
        s_offset.data[iY, iX] = std
    return s_offset


@deprecated(since="0.18.0", removal="0.19.0")
def _make_radius_vs_angle_model(
    signal,
    radial_signal_span,
    angleN=15,
    centre_x=None,
    centre_y=None,
    prepeak_range=None,
    show_progressbar=True,
):
    s_ra = signal.angular_slice_radial_average(
        angleN=angleN,
        centre_x=centre_x,
        centre_y=centre_y,
        show_progressbar=show_progressbar,
    )
    s_ra = s_ra.isig[radial_signal_span[0] : radial_signal_span[1]]

    m_ra = s_ra.create_model()

    # Fit 1 order polynomial to edges of data to account for the background
    sa = m_ra.axes_manager.signal_axes[0]
    if prepeak_range is None:
        prepeak_range = (sa.low_value, sa.index2value(4))
    m_ra.set_signal_range(prepeak_range[0], prepeak_range[1])
    m_ra.add_signal_range(sa.index2value(-3), sa.high_value)

    polynomial = Polynomial(1)
    m_ra.append(polynomial)
    m_ra.multifit(show_progressbar=show_progressbar)
    polynomial.set_parameters_not_free()

    m_ra.reset_signal_range()

    # Fit Gaussian to diffraction ring
    centre_initial = (sa.high_value + sa.low_value) * 0.5
    sigma = 3
    A = (s_ra - s_ra.min(axis=1)).data.max() * 2 * sigma

    gaussian = Gaussian(A=A, sigma=sigma, centre=centre_initial)
    gaussian.centre.bmin = sa.low_value
    gaussian.centre.bmax = sa.high_value
    m_ra.append(gaussian)
    return m_ra


@deprecated(since="0.18.0", removal="0.19.0")
def get_radius_vs_angle(
    signal,
    radial_signal_span,
    angleN=15,
    centre_x=None,
    centre_y=None,
    prepeak_range=None,
    show_progressbar=True,
):
    """
    Get radius of a ring as a function of angle.

    This is done by radially averaging angular slices of the image,
    followed by firstly fitting the background around the specified ring,
    then fitting a Gaussian to the remaining intensity in the ring (which
    gets reduced to a peak due to the radial average).

    Useful for finding if a ring is circular or elliptical.
    The centre position has be to set using the offset parameter in
    signal.axes_manager.

    Parameters
    ----------
    signal : HyperSpy 2D signal
        Diffraction image with calibrated centre point.
        This can be set manually, by looking at the diffraction pattern,
        or use a function like radial.get_optimal_centre_position
    radial_signal_span : tuple
    angleN : default 15
    centre_x, centre_y : NumPy arrays
    prepeak_range : tuple, optional
    show_progressbar : default True

    Returns
    -------
    s_centre : HyperSpy 1D signal

    Examples
    --------
    >>> s = pxm.data.dummy_data.get_single_ring_diffraction_signal()
    >>> s.axes_manager.signal_axes[0].offset = -105
    >>> s.axes_manager.signal_axes[1].offset = -67
    >>> import pyxem.utils.radial_utils as ra
    >>> s_centre = ra.get_radius_vs_angle(s, (35, 45), show_progressbar=False)

    """
    m_ra = _make_radius_vs_angle_model(
        signal=signal,
        radial_signal_span=radial_signal_span,
        angleN=angleN,
        centre_x=centre_x,
        centre_y=centre_y,
        prepeak_range=prepeak_range,
        show_progressbar=show_progressbar,
    )
    m_ra.multifit(optimizer="lm", bounded=True, show_progressbar=show_progressbar)

    s_centre = m_ra.components.Gaussian.centre.as_signal()
    return s_centre


@deprecated(since="0.18.0", removal="0.19.0")
def get_angle_image_comparison(s0, s1, angleN=12, mask_radius=None):
    """Compare two images by overlaying one on the other in slices.

    This function takes two images, extracts different angular slices
    and combines them into one image.

    Useful for comparing two diffraction images, to see if the rings
    have the same radius.

    Parameters
    ----------
    s0, s1 : HyperSpy 2D Signal
        Both signals need to have the same shape, and no navigation
        dimensions.
    angleN : int, default 12
        Number of angular slices.
    mask_radius : int, optional
        Mask the centre of the image. The default is not to mask anything.
        Useful to mask the most intense parts of the diffraction pattern,
        so the less intense parts can be visualized.

    Returns
    -------
    comparison_signal : HyperSpy 2D

    Examples
    --------
    >>> from pyxem.data.dummy_data import MakeTestData
    >>> test_data0 = MakeTestData(300, 300)
    >>> test_data0.add_ring(150, 150, 40)
    >>> test_data1 = MakeTestData(300, 300)
    >>> test_data1.add_ring(150, 150, 60)
    >>> s0 = test_data0.signal
    >>> s1 = test_data1.signal
    >>> s0.axes_manager[0].offset, s0.axes_manager[1].offset = -150, -150
    >>> s1.axes_manager[0].offset, s1.axes_manager[1].offset = -150, -150
    >>> import pyxem.utils.radial_utils as ra
    >>> s = ra.get_angle_image_comparison(s0, s1)
    >>> s.plot()

    Mask the inner parts

    >>> s = ra.get_angle_image_comparison(s0, s1, mask_radius=10)

    """
    if s0.axes_manager.shape != s1.axes_manager.shape:
        raise ValueError("s0 and s1 need to have the same shape")
    s = s0.deepcopy()
    angle_array = np.ogrid[0 : 2 * np.pi : (1 + angleN) * 1j]
    for i in range(len(angle_array[:-1])):
        if i % 2:
            angle0, angle1 = angle_array[i : i + 2]
            bool_array = pst._get_angle_sector_mask(s, angle0, angle1)
            s.data[bool_array] = s1.data[bool_array]

    if mask_radius is not None:
        am = s.axes_manager
        mask = pst._make_circular_mask(
            am[0].value2index(0.0),
            am[1].value2index(0.0),
            am[0].size,
            am[1].size,
            mask_radius,
        )
        mask = np.invert(mask)
        s.data *= mask
    return s


@deprecated(since="0.18.0", removal="0.19.0")
def _get_holz_angle(electron_wavelength, lattice_parameter):
    """
    Parameters
    ----------
    electron_wavelength : scalar
        In nanometers
    lattice_parameter : scalar
        In nanometers

    Returns
    -------
    scattering_angle : scalar
        Scattering angle in radians

    Examples
    --------
    >>> import pyxem.utils.radial_utils as ra
    >>> lattice_size = 0.3905 # STO-(001) in nm
    >>> wavelength = 2.51/1000 # Electron wavelength for 200 kV
    >>> angle = ra._get_holz_angle(wavelength, lattice_size)

    """
    k0 = 1.0 / electron_wavelength
    kz = 1.0 / lattice_parameter
    in_root = kz * ((2 * k0) - kz)
    sin_angle = (in_root**0.5) / k0
    angle = math.asin(sin_angle)
    return angle


@deprecated(since="0.18.0", removal="0.19.0")
def _scattering_angle_to_lattice_parameter(electron_wavelength, angle):
    """Convert scattering angle data to lattice parameter sizes.

    Parameters
    ----------
    electron_wavelength : float
        Wavelength of the electrons in the electron beam. In nm.
        For 200 kV electrons: 0.00251 (nm)
    angle : NumPy array
        Scattering angle, in radians.

    Returns
    -------
    lattice_parameter : NumPy array
        Lattice parameter, in nanometers

    Examples
    --------
    >>> import pyxem.utils.radial_utils as ra
    >>> angle_list = [0.1, 0.1, 0.1, 0.1] # in radians
    >>> wavelength = 2.51/1000 # Electron wavelength for 200 kV
    >>> lattice_size = ra._scattering_angle_to_lattice_parameter(
    ...     wavelength, angle_list)

    """

    k0 = 1.0 / electron_wavelength
    kz = k0 - (k0 * ((1 - (np.sin(angle) ** 2)) ** 0.5))
    return 1 / kz


@deprecated(since="0.18.0", removal="0.19.0")
def _get_xy_points_from_radius_angle_plot(s_ra):
    x_list = []
    y_list = []
    for angle, radius in zip(s_ra.axes_manager[0].axis, s_ra.data):
        dx = -math.cos(angle) * radius
        dy = -math.sin(angle) * radius
        x_list.append(dx)
        y_list.append(dy)
    x_list = np.array(x_list)
    y_list = np.array(y_list)
    return (x_list, y_list)


@deprecated(since="0.18.0", removal="0.19.0")
def _fit_ellipse_to_xy_points(x, y):
    """Fit an ellipse to a list of x and y points.

    Parameters
    ----------
    x, y : NumPy 2D array

    Returns
    -------
    ellipse_parameters : NumPy array

    """
    xx = x * x
    yy = y * y
    xy = x * y
    ones = np.ones_like(x)
    D = np.vstack((xx, xy, yy, x, y, ones))
    S = np.dot(D, D.T)
    C = np.zeros((6, 6))
    C[0, 2], C[2, 0] = 2, 2
    C[1, 1] = -1
    A, B = la.eig(np.dot(la.inv(S), C))
    i = np.argmax(np.abs(A))
    g = B[:, i]
    return g


@deprecated(since="0.18.0", removal="0.19.0")
def _get_ellipse_parameters(g):
    """
    Parameters
    ----------
    g : NumPy array

    Returns
    -------
    xC, yC : floats
        Centre position
    semi_len0, semi_len1 : floats
        Length of minor and major semi-axis
    rot : float
        Angle between semi_len0 and the positive x-axis. Since semi_len0,
        is not necessarily the longest semi-axis, the rotation will _not_ be
        between the major semi-axis and the positive x-axis.
        In radians, between 0 and pi. The rotation is clockwise, so
        at rot = 0.1 the ellipse will be pointing in the positive x-direction,
        and negative y-direction.
    eccen : float
        Eccentricity of the ellipse

    Note
    ----
    http://mathworld.wolfram.com/Ellipse.html

    """
    a, b, c, d, f, g = g[0], g[1] / 2, g[2], g[3] / 2, g[4] / 2, g[5]
    b2ac = b * b - a * c
    xC = (c * d - b * f) / b2ac
    yC = (a * f - b * d) / b2ac
    frac_top = 2 * (a * f * f + c * d * d + g * b * b - 2 * b * d * f - a * c * g)
    frac_square_bot = (a - c) * (a - c) + 4 * b * b
    semi_len0 = math.sqrt(frac_top / (b2ac * (math.sqrt(frac_square_bot) - (a + c))))
    semi_len1 = math.sqrt(frac_top / (b2ac * (-math.sqrt(frac_square_bot) - (a + c))))
    if b == 0:
        if a < c:
            rot = 0
        else:
            rot = math.pi * 0.5
    else:
        if a < c:
            rot = math.atan((2 * b) / (a - c)) / 2
        else:
            rot = math.pi / 2 + (math.atan((2 * b) / (a - c)) / 2)
    rot = rot % np.pi
    eccen = math.sqrt(1 - ((b * b) / (a * a)))
    return (xC, yC, semi_len0, semi_len1, rot, eccen)


@deprecated(since="0.18.0", removal="0.19.0")
def _get_marker_list(ellipse_parameters, x_list=None, y_list=None):
    xC, yC, semi_len0, semi_len1, rot, ecce = _get_ellipse_parameters(
        ellipse_parameters
    )
    marker_list = []
    if x_list is not None:
        offsets = np.array([x_list, y_list]).T
        marker_list.append(hs.plot.markers.Points(offsets=offsets))
    from pyxem.utils.ransac_ellipse_tools import ellipse_to_markers

    marker_list.append(
        ellipse_to_markers(ellipse_array=[xC, yC, semi_len0, semi_len1, rot])
    )
    return marker_list


@deprecated(since="0.18.0", removal="0.19.0")
def fit_single_ellipse_to_signal(
    s, radial_signal_span, prepeak_range=None, angleN=20, show_progressbar=True
):
    """
    Parameters
    ----------
    s : HyperSpy Signal2D
    radial_signal_span : tuple
    prepeak_range : tuple, optional
    angleN : int, default 20
    show_progressbar : bool, default True

    Returns
    -------
    signal : HyperSpy 2D signal
        Fitted ellipse and fitting points are stored as HyperSpy markers
        in the metadata
    xC, yC : floats
        Centre position
    semi_len0, semi_len1 : floats
        Length of minor and major semi-axis
    rot : float
        Angle between semi_len0 and the positive x-axis. Since semi_len0,
        is not necessarily the longest semi-axis, the rotation will _not_ be
        between the major semi-axis and the positive x-axis.
        In radians, between 0 and pi. The rotation is clockwise, so
        at rot = 0.1 the ellipse will be pointing in the positive x-direction,
        and negative y-direction.
    eccen : float
        Eccentricity of the ellipse

    signal, xC, yC, semi0, semi1, rot, ecc

    Examples
    --------
    >>> from pyxem.data.dummy_data import make_diffraction_test_data as mdtd
    >>> s = pxm.signals.Diffraction2D(np.zeros((200, 220)))
    >>> s.axes_manager[0].offset, s.axes_manager[1].offset = -100, -110
    >>> xx, yy = np.meshgrid(s.axes_manager[0].axis, s.axes_manager[1].axis)
    >>> ellipse_ring = mdtd._get_elliptical_ring(xx, yy, 0, 0, 50, 70, 0.8)
    >>> s.data += ellipse_ring
    >>> from pyxem.utils.radial_utils import fit_single_ellipse_to_signal
    >>> output = fit_single_ellipse_to_signal(
    ...     s, (40, 80), angleN=30, show_progressbar=False)

    """
    s_ra = get_radius_vs_angle(
        s,
        radial_signal_span,
        angleN=angleN,
        prepeak_range=prepeak_range,
        show_progressbar=show_progressbar,
    )
    x, y = _get_xy_points_from_radius_angle_plot(s_ra)
    ellipse_parameters = _fit_ellipse_to_xy_points(x, y)
    xC, yC, semi0, semi1, rot, ecc = _get_ellipse_parameters(ellipse_parameters)
    marker_list = _get_marker_list(ellipse_parameters, x_list=x, y_list=y)
    s_m = s.deepcopy()
    s_m.add_marker(marker_list, permanent=True, plot_marker=False)
    return s_m, xC, yC, semi0, semi1, rot, ecc


@deprecated(since="0.18.0", removal="0.19.0")
def fit_ellipses_to_signal(
    s, radial_signal_span_list, prepeak_range=None, angleN=20, show_progressbar=True
):
    """
    Parameters
    ----------
    s : HyperSpy Signal2D
    radial_signal_span_list : tuple
    prepeak_range : tuple, optional
    angleN : list or int, default 20
    show_progressbar : bool, default True

    Returns
    -------
    signal : HyperSpy 2D signal
        Fitted ellipse and fitting points are stored as HyperSpy markers
        in the metadata
    xC, yC : floats
        Centre position
    semi_len0, semi_len1 : floats
        Length of the two semi-axes.
    rot : float
        Angle between semi_len0 and the positive x-axis. Since semi_len0,
        is not necessarily the longest semi-axis, the rotation will _not_ be
        between the major semi-axis and the positive x-axis.
        In radians, between 0 and pi. The rotation is clockwise, so
        at rot = 0.1 the ellipse will be pointing in the positive x-direction,
        and negative y-direction.
    eccen : float
        Eccentricity of the ellipse

    Examples
    --------
    >>> from pyxem.data.dummy_data import make_diffraction_test_data as mdtd
    >>> s = pxm.signals.Diffraction2D(np.zeros((200, 220)))
    >>> s.axes_manager[0].offset, s.axes_manager[1].offset = -100, -110
    >>> xx, yy = np.meshgrid(s.axes_manager[0].axis, s.axes_manager[1].axis)
    >>> ellipse_ring = mdtd._get_elliptical_ring(xx, yy, 0, 0, 50, 70, 0.8)
    >>> s.data += ellipse_ring
    >>> from pyxem.utils.radial_utils import fit_ellipses_to_signal
    >>> output = fit_ellipses_to_signal(
    ...     s, [(40, 80)], angleN=30, show_progressbar=False)

    """
    if not isiterable(angleN):
        angleN = [angleN] * len(radial_signal_span_list)
    else:
        if len(angleN) != len(radial_signal_span_list):
            raise ValueError(
                "angleN and radial_signal_span_list needs to have " "the same length"
            )
    marker_list = []
    ellipse_list = []
    for i, (radial_signal_span, aN) in enumerate(zip(radial_signal_span_list, angleN)):
        s_ra = get_radius_vs_angle(
            s,
            radial_signal_span,
            angleN=aN,
            prepeak_range=prepeak_range,
            show_progressbar=show_progressbar,
        )
        x, y = _get_xy_points_from_radius_angle_plot(s_ra)
        ellipse_parameters = _fit_ellipse_to_xy_points(x, y)
        output = _get_ellipse_parameters(ellipse_parameters)
        ellipse_list.append(output)
        marker_list.extend(
            _get_marker_list(
                ellipse_parameters,
                x_list=x,
                y_list=y,
            )
        )
    s_m = s.deepcopy()
    s_m.add_marker(marker_list, permanent=True, plot_marker=False)
    return s_m, ellipse_list



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

from functools import cached_property, partial
from warnings import warn

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from scipy.spatial import distance_matrix
import dask.array as da

from hyperspy.signals import BaseSignal, Signal1D
from hyperspy._signals.lazy import LazySignal
from hyperspy.drawing._markers.points import Points
from hyperspy.misc.utils import isiterable

from pyxem.utils._signals import (
    _transfer_navigation_axes_to_signal_axes,
)
from pyxem.utils.vectors import (
    detector_to_fourier,
    get_npeaks,
    filter_vectors_ragged,
    filter_vectors_edge_ragged,
    filter_vectors_near_basis,
    _reverse_pos,
    cluster,
    vectors_to_polar,
)

from pyxem.utils._slicers import Slicer

from pyxem.utils._subpixel_finding import (
    _conventional_xc_map,
    _center_of_mass_map,
    _get_simulated_disc,
    _wrap_columns,
)

from pyxem.utils._deprecated import deprecated

"""
Signal class for diffraction vectors.

There are two cases that are supported:

1. A map of diffraction vectors, which will in general be a ragged signal of
signals. It the navigation dimensions of the map and contains a signal for each
peak at every position.

2. A list of diffraction vectors with dimensions < n | 2 > where n is the
number of peaks.
"""


class DiffractionVectors(BaseSignal):
    """Class for diffraction vectors in reciprocal space.

    Diffraction vectors are defined as the vectors from the center of the
    diffraction pattern to the diffraction peaks. Note the DiffractionVectors
    class is likely to change as the API for diffraction vectors is developed.

    Attributes
    ----------
    cartesian : numpy.array
        Array of 3-vectors describing Cartesian coordinates associated with
        each diffraction vector.
    hkls : numpy.array
        Array of Miller indices associated with each diffraction vector
        following indexation.
    """

    _signal_dimension = 0
    _signal_type = "diffraction_vectors"

    def __init__(self, *args, **kwargs):
        _scales = kwargs.pop("scales", None)
        _offsets = kwargs.pop("offsets", None)
        _detector_shape = kwargs.pop("detector_shape", None)
        _column_names = kwargs.pop("column_names", None)
        _units = kwargs.pop("units", None)
        super().__init__(*args, **kwargs)
        self._set_up_vector(_scales, _offsets, _detector_shape, _column_names, _units)
        if (
            self._is_object_dtype is None
        ):  # empty signal with data=None due to `_deepcopy_with_new_data`
            pass
        elif self._is_object_dtype:
            self.ragged = True
        elif self.ragged == True:
            self.ragged = False

    def _repr_html_(self):
        table = '<table align="center">'
        table += f'<tr><td colspan= "{self.num_columns + 1}"> <center> {self.__str__()[1:-1]} </center></td></tr>'
        table += (
            f'<tr><td colspan= "{self.num_columns + 1}"> <center> Current Index:{self.axes_manager.indices} '
            f"</center> </td></tr>"
        )
        table += f'<tr><td colspan= "{self.num_columns + 1}">  </td></tr>'

        table += "<tr><th>column_names:</th>"
        for col in self.column_names:
            table += f"<th><center>{col}</center></th>"
        table += "</tr>"
        table += "<th>units:</th>"
        for col in self.units:
            table += f"<th><center>{col}</center></th>"
        table += "</tr>"

        vectors = self._get_current_data()
        if vectors.dtype.kind == "O":
            vectors = vectors[0]
        if self.num_columns == 1:
            vectors = np.array([vectors]).T
        if vectors.ndim == 1:
            vectors = np.array([vectors])
        for i, row in enumerate(vectors):
            table += "<tr>"
            table += f"<td><center>{i}</center></td>"
            for col in row:
                table += f"<td><center>{col}</center></td>"
            table += "</tr>"
            if i > 10:
                table += f'<tr><td colspan= "{self.num_columns + 1}"> ... </td></tr>'
                break
        table += "</table>"
        return table

    def _set_up_vector(
        self,
        scales=None,
        offsets=None,
        detector_shape=None,
        column_names=None,
        units=None,
    ):
        self.metadata.add_node("VectorMetadata")
        if scales is not None or "scales" not in self.metadata.VectorMetadata:
            self.metadata.VectorMetadata["scales"] = scales
        if offsets is not None or "offsets" not in self.metadata.VectorMetadata:
            self.metadata.VectorMetadata["offsets"] = offsets
        if (
            detector_shape is not None
            or "detector_shape" not in self.metadata.VectorMetadata
        ):
            self.metadata.VectorMetadata["detector_shape"] = detector_shape

        if (
            column_names is not None
            or "column_names" not in self.metadata.VectorMetadata
        ):
            self.metadata.VectorMetadata["column_names"] = column_names
        if units is not None or "units" not in self.metadata.VectorMetadata:
            self.metadata.VectorMetadata["units"] = units

        self.cartesian = None
        self.hkls = None
        self.is_real_units = False
        self.has_intensity = False

        self.ivec = Slicer(self)

    @classmethod
    def from_peaks(
        cls, peaks, center=None, calibration=None, column_names=None, units=None
    ):
        """Takes a list of peak positions (pixel coordinates) and returns
        an instance of `Diffraction2D`

        Parameters
        ----------
        peaks : hyperspy.api.signals.BaseSignal
            Signal containing lists (np.array) of pixel coordinates specifying
            the reflection positions
        center : np.array or None
            Diffraction pattern center in array indices.
        calibration : np.array or None
            Calibration in reciprocal Angstroms per pixels for each of the dimensions.

        Returns
        -------
        vectors : :class:`pyxem.signals.diffraction_vectors.DiffractionVectors`
            List of diffraction vectors
        """
        if center is None and peaks.metadata.has_item("Peaks.signal_axes"):
            center = [
                ax.offset / ax.scale for ax in peaks.metadata.Peaks.signal_axes[::-1]
            ]
        elif center is not None:
            center = -np.array(center)
        else:
            raise ValueError(
                "A center and calibration must be provided unless the"
                "peaks.metadata.Peaks.signal_axes is set."
            )
        if calibration is None and peaks.metadata.has_item("Peaks.signal_axes"):
            calibration = [ax.scale for ax in peaks.metadata.Peaks.signal_axes[::-1]]
        elif calibration is not None:
            pass  # calibration is already set
        else:
            raise ValueError(
                "A center and calibration must be provided unless the"
                "peaks.metadata.Peaks.signal_axes is set."
            )

        if column_names is None and peaks.metadata.has_item("Peaks.signal_axes"):
            column_names = [
                str(ax.name) for ax in peaks.metadata.Peaks.signal_axes[::-1]
            ]
        elif column_names is not None:
            pass
        else:
            column_names = ["x", "y"]

        if units is None and peaks.metadata.has_item("Peaks.signal_axes"):
            units = [str(ax.units) for ax in peaks.metadata.Peaks.signal_axes[::-1]]
        elif units is not None:
            pass
        else:
            units = ["", ""]

        if not isiterable(calibration):
            calibration = [
                calibration,
                calibration,
            ]  # same calibration for both dimensions

        if isinstance(peaks, LazySignal):
            num_cols = peaks.data[(0,) * peaks.data.ndim].compute().shape[1]
        else:
            num_cols = peaks.data[(0,) * peaks.data.ndim].shape[1]
        if num_cols == len(calibration) + 1:
            # account for the intensity column
            center = list(center) + [
                0,
            ]
            calibration = list(calibration) + [
                1,
            ]
            has_intensity = True
        else:
            has_intensity = False

        if num_cols == len(column_names) + 1:
            column_names = list(column_names) + ["intensity"]
            units = list(units) + ["a.u."]

        vectors = peaks.map(
            lambda x, cen, cal: (x + cen) * cal,
            cal=calibration,
            cen=center,
            inplace=False,
            ragged=True,
            output_signal_size=(),
            output_dtype=object,
        )
        vectors.set_signal_type("diffraction_vectors")
        if isinstance(peaks, LazySignal):
            vectors = vectors.as_lazy()
        vectors._set_up_vector(
            scales=calibration, column_names=column_names, units=units
        )
        vectors.center = center
        vectors.has_intensity = has_intensity
        vectors.column_names = column_names
        return vectors

    def subpixel_refine(
        self,
        signal,
        method="center-of-mass",
        disk_r=None,
        upsample_factor=2,
        square_size=10,
        columns=None,
        **kwargs,
    ):
        """
        Refine the positions of the diffraction vectors using subpixel
        interpolation.

        Parameters
        ----------
        signal : hyperspy.api.signals.BaseSignal
            The signal which will be used to refine the diffraction vectors.
        method : str
            The method used to refine the diffraction vectors. Currently
            supported methods are 'cross-correlation' and
            and "center-of-mass".
        disk_r : int or None
            The radius of the disk used for the cross-correlation method in pixels.
        upsample_factor : int
            The upsample factor used for the cross-correlation method.
        square_size : int
            The size of the square used for both the center-of-mass and cross-correlation methods.
        columns : list
            The columns for the pixels of interest. If None, the first two columns are used.
        kwargs : dict
            Additional keyword arguments to be passed to the map method.

        Returns
        -------
        refined_vectors : DiffractionVectors
            The refined vectors.
        """
        if columns is None:
            columns = [0, 1]
        method_dict = {
            "cross-correlation": _conventional_xc_map,
            "center-of-mass": _center_of_mass_map,
        }
        if method not in method_dict:
            raise ValueError(
                f"The method parameter must be one of {list(method_dict.keys())}"
            )
        if method == "cross-correlation":
            kwargs["upsample_factor"] = upsample_factor
            kwargs["kernel"] = _get_simulated_disc(square_size, disk_r)
        kwargs["square_size"] = square_size

        signal_axes = signal.axes_manager.signal_axes
        offsets = np.array([ax.offset for ax in signal_axes])
        scales = np.array([ax.scale for ax in signal_axes])

        funct = method_dict[method]
        pixels = self.get_pixel_vectors(
            offsets=offsets,
            scales=scales,
            shape=signal.axes_manager._signal_shape_in_array,
            square_size=square_size,
            columns=columns,
        )
        method_func = partial(_wrap_columns, f=funct, columns=columns)

        _wrap_columns
        refined_vectors = signal.map(
            method_func,
            vectors=pixels,
            inplace=False,
            ragged=True,
            offsets=offsets,
            scales=scales,
            columns=columns,
            silence_warnings=True,
            **kwargs,
        )
        refined_vectors.set_signal_type("diffraction_vectors")
        refined_vectors._set_up_vector(
            scales=self.scales, column_names=self.column_names
        )
        return refined_vectors

    @property
    def pixel_vectors(self):
        return self.get_pixel_vectors()

    def get_pixel_vectors(
        self, offsets=None, scales=None, square_size=None, shape=None, columns=None
    ):
        """Returns the diffraction vectors in pixel coordinates."""
        if offsets is None:
            offsets = self.offsets
        if scales is None:
            scales = self.scales

        def get_pixels(x, off, scale, square_size=None, shape=None, columns=None):
            if columns is not None:
                pixels = np.round(
                    (x[:, columns] - off[columns]) / scale[columns]
                ).astype(int)
            else:
                pixels = np.round((x - off) / scale).astype(int)

            if columns is not None:
                num_cols = x.shape[1]
                other_indexes = np.arange(num_cols)[
                    np.logical_not(np.isin(np.arange(num_cols), columns))
                ]
                other_columns = x[:, other_indexes]
            if square_size is not None and shape is not None:
                is_in = np.all(pixels > (square_size / 2) + 1, axis=1) * np.all(
                    np.array(shape) - pixels > (square_size / 2) + 1, axis=1
                )
                pixels = pixels[is_in]
                if columns is not None:
                    other_columns = other_columns[is_in]
            if columns is not None:
                pixels = np.hstack([pixels, other_columns])
            return pixels

        pixels = self.map(
            get_pixels,
            off=offsets,
            scale=scales,
            square_size=square_size,
            shape=shape,
            inplace=False,
            ragged=True,
            columns=columns,
            silence_warnings=True,
        )
        return pixels

    @property
    def _is_object_dtype(self):
        try:
            if self.data[0] is None:
                return None
            else:
                return self.data.dtype.kind == "O"
        except IndexError:
            return None

    @cached_property
    def num_columns(self):
        if self._is_object_dtype:
            if isinstance(self.data, da.Array):
                shape = self.data[self.data.ndim * (0,)].compute().shape
            else:
                shape = self.data[self.data.ndim * (0,)].shape
            if shape is None:
                return 0
            elif len(shape) == 1:
                return 1
            else:
                return shape[1]
        else:
            return self.data.shape[-1]

    @property
    def units(self):
        if self.metadata.VectorMetadata["units"] is None:
            return [
                None,
            ] * self.num_columns
        else:
            return self.metadata.VectorMetadata["units"]

    @units.setter
    def units(self, value):
        if isinstance(value, str) and self.num_columns == 1:
            value = [value]
        if (
            isiterable(value)
            and len(value) == self.num_columns
            and not isinstance(value, str)
        ):
            self.metadata.VectorMetadata["units"] = value

        elif isiterable(value) and len(value) != self.num_columns:
            raise ValueError(
                "The len of the units parameter must equal the number of"
                " columns in the underlying vector data."
            )
        else:
            self.metadata.VectorMetadata["units"] = [
                value,
            ] * self.num_columns

    @property
    def scales(self):
        return self.metadata.VectorMetadata["scales"]

    @scales.setter
    def scales(self, value):
        if isiterable(value) and len(value) == self.num_columns:
            self.metadata.VectorMetadata["scales"] = value
        elif isiterable(value) and len(value) != self.num_columns:
            raise ValueError(
                "The len of the scales parameter must equal the number of"
                " columns in the underlying vector data."
            )
        else:
            self.metadata.VectorMetadata["scales"] = [
                value,
            ] * self.num_columns

    @property
    def column_names(self):
        if self.metadata.VectorMetadata["column_names"] is None:
            return [
                None,
            ] * self.num_columns
        else:
            return self.metadata.VectorMetadata["column_names"]

    @column_names.setter
    def column_names(self, value):
        if value is None:
            value = [f"column_{i}" for i in range(self.num_columns)]

        if isinstance(value, str):
            value = [value]
        if len(value) != self.num_columns:
            raise ValueError(
                f"The len of the column_names parameter: {len(value)} must equal the"
                f" number of columns in the underlying vector data: {self.num_columns}."
            )

        self.metadata.VectorMetadata["column_names"] = value

    @property
    def offsets(self):
        return self.metadata.VectorMetadata["offsets"]

    @offsets.setter
    def offsets(self, value):
        if isiterable(value) and len(value) == self.num_columns:
            self.metadata.VectorMetadata["offsets"] = np.array(value)

        elif isiterable(value) and len(value) != self.num_columns:
            raise ValueError(
                "The len of the scales parameter must equal the number of"
                "columns in the underlying vector data."
            )
        else:
            self.metadata.VectorMetadata["offsets"] = [
                value,
            ] * self.num_columns

    def __lt__(self, other):
        if self.ragged:
            kwargs = dict(output_signal_size=(), output_dtype=object)
        else:
            kwargs = dict()
        return self.map(
            lambda x, other: x < other,
            other=other,
            inplace=False,
            ragged=self.ragged,
            silence_warnings=True,
            **kwargs,
        )

    def __le__(self, other):
        if self.ragged:
            kwargs = dict(output_signal_size=(), output_dtype=object)
        else:
            kwargs = dict()
        return self.map(
            lambda x, other: x <= other,
            other=other,
            inplace=False,
            ragged=self.ragged,
            silence_warnings=True,
            **kwargs,
        )

    def __gt__(self, other):
        if self.ragged:
            kwargs = dict(output_signal_size=(), output_dtype=object)
        else:
            kwargs = dict()
        return self.map(
            lambda x, other: x > other,
            other=other,
            inplace=False,
            ragged=self.ragged,
            silence_warnings=True,
            **kwargs,
        )

    def __ge__(self, other):
        if self.ragged:
            kwargs = dict(output_signal_size=(), output_dtype=object)
        else:
            kwargs = dict()
        return self.map(
            lambda x, other: x >= other,
            other=other,
            inplace=False,
            ragged=self.ragged,
            silence_warnings=True,
            **kwargs,
        )

    @property
    def center(self):
        """The center of the diffraction pattern in pixels."""
        return np.divide(self.offsets, self.scales)

    @center.setter
    def center(self, value):
        self.offsets = np.multiply(value, self.scales)

    @property
    @deprecated(
        since="0.15",
        alternative="pyxem.signals.DiffractionVectors.scales",
        removal="1.0.0",
    )
    def pixel_calibration(self):
        return self.scales

    @pixel_calibration.setter
    @deprecated(
        since="0.15",
        alternative="pyxem.signals.DiffractionVectors.scales",
        removal="1.0.0",
    )
    def pixel_calibration(self, value):
        self.scales = value

    @property
    def detector_shape(self):
        return self.metadata.VectorMetadata["detector_shape"]

    @detector_shape.setter
    def detector_shape(self, value):
        self.metadata.VectorMetadata["detector_shape"] = value

    def _get_navigation_positions(self, flatten=False, real_units=True):
        nav_indexes = np.array(
            list(np.ndindex(self.axes_manager._navigation_shape_in_array))
        )
        if not real_units:
            scales = [1 for a in self.axes_manager.navigation_axes]
            offsets = [0 for a in self.axes_manager.navigation_axes]
        else:
            scales = [a.scale for a in self.axes_manager.navigation_axes[::-1]]
            offsets = [a.offset for a in self.axes_manager.navigation_axes[::-1]]

        if flatten:
            real_nav = np.array(
                [
                    np.array(ind) * scales + offsets
                    for ind in np.array(list(nav_indexes))
                ]
            )
        else:
            real_nav = np.reshape(
                [
                    np.array(ind) * scales + offsets
                    for ind in np.array(list(nav_indexes))
                ],
                self.axes_manager._navigation_shape_in_array + (-1,),
            )
        return real_nav

    @property
    def num_rows(self):
        if self._is_object_dtype:
            return None
        elif len(self.axes_manager.signal_axes) == 1:
            return 1
        else:
            return self.data.shape[-2]

    def flatten_diffraction_vectors(
        self,
        real_units=True,
    ):
        """Flattens the diffraction vectors into a `DiffractionVector2D` object.

        Each navigation axis is transformed into a vector defined by the scale and offset.
        This method allows purely vector based actions like filtering or determining unique
        values.

        Parameters
        ----------
        real_units: bool
            If the navigation dimension should be flattened based on the pixel position
            or the real value as determined by the scale and offset.
        """
        from pyxem.signals.diffraction_vectors2d import DiffractionVectors2D

        if self.axes_manager._navigation_shape_in_array == ():
            return self

        if self._is_object_dtype:
            nav_positions = self._get_navigation_positions(
                flatten=False, real_units=real_units
            )
            if self.num_columns == 1:
                vectors = np.vstack(
                    [
                        np.hstack(
                            [
                                np.tile(
                                    nav_positions[ind][::-1], (len(self.data[ind]), 1)
                                ),
                                self.data[ind][:, np.newaxis],
                            ]
                        )
                        for ind in np.ndindex(
                            self.axes_manager._navigation_shape_in_array
                        )
                    ]
                )
            else:
                vectors = np.vstack(
                    [
                        np.hstack(
                            [
                                np.tile(
                                    nav_positions[ind][::-1], (len(self.data[ind]), 1)
                                ),
                                self.data[ind],
                            ]
                        )
                        for ind in np.ndindex(
                            self.axes_manager._navigation_shape_in_array
                        )
                    ]
                )
        else:
            nav_positions = self._get_navigation_positions(
                flatten=True, real_units=real_units
            )
            navs = np.repeat(nav_positions, self.num_rows, axis=0)
            data = self.data.reshape((-1, self.num_columns))
            vectors = np.vstack((navs, data))
        if real_units:
            scales = [a.scale for a in self.axes_manager.navigation_axes]
            offsets = [a.offset for a in self.axes_manager.navigation_axes]
        else:
            scales = [1 for a in self.axes_manager.navigation_axes]
            offsets = [0 for a in self.axes_manager.navigation_axes]

        if self.offsets is None:
            column_offsets = [
                None,
            ] * (vectors.shape[1] - len(self.axes_manager.navigation_axes))
        else:
            column_offsets = self.offsets

        if self.scales is None:
            column_scale = [
                None,
            ] * (vectors.shape[1] - len(self.axes_manager.navigation_axes))
        else:
            column_scale = self.scales

        column_offsets = np.append(column_offsets, offsets)
        column_scale = np.append(column_scale, scales)

        column_names = np.append(
            [a.name for a in self.axes_manager.navigation_axes], self.column_names
        )

        if real_units:
            units = np.append(
                [a.units for a in self.axes_manager.navigation_axes], self.units
            )
        else:
            units = np.append(
                ["pixels"] * len(self.axes_manager.navigation_axes), self.units
            )

        return DiffractionVectors2D(
            vectors,
            column_offsets=column_offsets,
            column_scale=column_scale,
            units=units,
            column_names=column_names,
        )

    def plot_diffraction_vectors(
        self,
        xlim=1.0,
        ylim=1.0,
        unique_vectors=None,
        distance_threshold=0.01,
        method="distance_comparison",
        min_samples=1,
        image_to_plot_on=None,
        image_cmap="gray",
        plot_label_colors=False,
        distance_threshold_all=0.005,
    ):  # pragma: no cover
        """Plot the unique diffraction vectors.

        Parameters
        ----------
        xlim : float
            The maximum x coordinate to be plotted.
        ylim : float
            The maximum y coordinate in reciprocal Angstroms to be plotted.
        unique_vectors : DiffractionVectors, optional
            The unique vectors to be plotted (optional). If not given, the
            unique vectors will be found by get_unique_vectors.
        distance_threshold : float, optional
            The minimum distance in reciprocal Angstroms between diffraction
            vectors for them to be considered unique diffraction vectors.
            Will be passed to get_unique_vectors if no unique vectors are
            given.
        method : str
            The method to use to determine unique vectors, if not given.
            Valid methods are 'strict', 'distance_comparison' and 'DBSCAN'.
            'strict' returns all vectors that are strictly unique and
            corresponds to distance_threshold=0.
            'distance_comparison' checks the distance between vectors to
            determine if some should belong to the same unique vector,
            and if so, the unique vector is iteratively updated to the
            average value.
            'DBSCAN' relies on the DBSCAN [1] clustering algorithm, and
            uses the Eucledian distance metric.
        min_samples : int, optional
            The minimum number of not identical vectors within one cluster
            for it to be considered a core sample, i.e. to not be considered
            noise. Will be passed to get_unique_vectors if no unique vectors
            are given. Only used if method=='DBSCAN'.
        image_to_plot_on : BaseSignal, optional
            If provided, the vectors will be plotted on top of this image.
            The image must be calibrated in terms of offset and scale.
        image_cmap : str, optional
            The colormap to plot the image in.
        plot_label_colors : bool, optional
            If True (default is False), also the vectors contained within each
            cluster will be plotted, with colors according to their
            cluster membership. If True, the unique vectors will be
            calculated by get_unique_vectors. Requires on method=='DBSCAN'.
        distance_threshold_all : float, optional
            The minimum distance, in calibrated units, between diffraction
            vectors inside one cluster for them to be plotted. Only used if
            plot_label_colors is True and requires method=='DBSCAN'.

        Returns
        -------
        fig : matplotlib figure
            The plot as a matplotlib figure.

        """
        fig = plt.figure()
        ax = fig.add_subplot(111)
        offset, scale = 0.0, 1.0
        if image_to_plot_on is not None:
            offset = image_to_plot_on.axes_manager[-1].offset
            scale = image_to_plot_on.axes_manager[-1].scale
            ax.imshow(image_to_plot_on, cmap=image_cmap)
        else:
            ax.set_xlim(-xlim, xlim)
            ax.set_ylim(ylim, -ylim)
            ax.set_aspect("equal")

        if plot_label_colors is True and method == "DBSCAN":
            clusters = self.get_unique_vectors(
                distance_threshold,
                method="DBSCAN",
                min_samples=min_samples,
                return_clusters=True,
            )[1]
            labs = clusters.labels_[clusters.core_sample_indices_]
            # Get all vectors from the clustering not considered noise
            cores = clusters.components_
            if cores.size == 0:
                warn(
                    "No clusters were found. Check parameters, or "
                    "use plot_label_colors=False."
                )
            else:
                peaks = DiffractionVectors(cores)
                peaks.transpose(signal_axes=1)
                # Since this original number of vectors can be huge, we
                # find a reduced number of vectors that should be plotted, by
                # running a new clustering on all the vectors not considered
                # noise, considering distance_threshold_all.
                peaks = peaks.get_unique_vectors(
                    distance_threshold_all, min_samples=1, return_clusters=False
                )
                peaks_all_len = peaks.data.shape[0]
                labels_to_plot = np.zeros(peaks_all_len)
                peaks_to_plot = np.zeros((peaks_all_len, 2))
                # Find the labels of each of the peaks to plot by referring back
                # to the list of labels for the original vectors.
                for n, peak in zip(np.arange(peaks_all_len), peaks):
                    index = distance_matrix([peak.data], cores).argmin()
                    peaks_to_plot[n] = cores[index]
                    labels_to_plot[n] = labs[index]
                # Assign a color value to each label, and shuffle these so that
                # adjacent clusters hopefully get distinct colors.
                cmap_lab = matplotlib.colormaps["gist_rainbow"]
                lab_values_shuffled = np.arange(np.max(labels_to_plot) + 1)
                np.random.shuffle(lab_values_shuffled)
                labels_steps = np.array(
                    list(map(lambda n: lab_values_shuffled[int(n)], labels_to_plot))
                )
                labels_steps = labels_steps / (np.max(labels_to_plot) + 1)
                # Plot all peaks
                for lab, peak in zip(labels_steps, peaks_to_plot):
                    ax.plot(
                        (peak[0] - offset) / scale,
                        (peak[1] - offset) / scale,
                        ".",
                        color=cmap_lab(lab),
                    )
        if unique_vectors is None:
            unique_vectors = self.get_unique_vectors(
                distance_threshold, method=method, min_samples=min_samples
            )
        # Plot the unique vectors
        ax.plot(
            (unique_vectors.data.T[1] - offset) / scale,
            (unique_vectors.data.T[0] - offset) / scale,
            "kx",
        )
        plt.tight_layout()
        plt.axis("off")
        return fig

    def to_markers(self, **kwargs):
        new = self.map(
            _reverse_pos,
            inplace=False,
            ragged=True,
            silence_warnings=True,
        )
        return Points(offsets=new.data.T, **kwargs)

    @deprecated(
        since="0.17.0",
        removal="1.0.0",
        alternative="pyxem.signals.DiffractionVectors.to_markers",
    )
    def plot_diffraction_vectors_on_signal(self, signal, *args, **kwargs):
        """Plot the diffraction vectors on a signal.

        Parameters
        ----------
        signal : ElectronDiffraction2D
            The ElectronDiffraction2D signal object on which to plot the peaks.
            This signal must have the same navigation dimensions as the peaks.
        *args :
            Arguments passed to signal.plot()
        **kwargs :
            Keyword arguments passed to signal.plot()
        """
        signal.plot(*args, **kwargs)
        marker = self.to_markers(
            color=[
                "red",
            ]
        )
        signal.add_marker(marker, plot_marker=True, permanent=False)

    def get_magnitudes(self, columns=None, *args, **kwargs):
        """Calculate the magnitude of diffraction vectors.

        Parameters
        ----------
        columns : list, optional
            The columns of the diffraction vectors to be used to calculate
            the magnitude. If not given, the first two columns will be used.
        *args:
            Arguments to be passed to map().
        **kwargs:
            Keyword arguments to map().

        Returns
        -------
        magnitudes : BaseSignal
            A signal with navigation dimensions as the original diffraction
            vectors containg an array of gvector magnitudes at each
            navigation position.

        """
        if columns is None:
            columns = [0, 1]

        def get_magnitude(x):
            return np.linalg.norm(x[:, columns], axis=-1)

        magnitudes = self.map(
            get_magnitude, inplace=False, silence_warnings=True, *args, **kwargs
        )

        return magnitudes

    def get_magnitude_histogram(self, bins, *args, **kwargs):
        """Obtain a histogram of gvector magnitudes.

        Parameters
        ----------
        bins : numpy array
            The bins to be used to generate the histogram.
        *args:
            Arguments to get_magnitudes().
        **kwargs:
            Keyword arguments to get_magnitudes().

        Returns
        -------
        ghis : Signal1D
            Histogram of gvector magnitudes.

        """
        gmags = self.get_magnitudes(*args, **kwargs)

        if len(self.axes_manager.signal_axes) == 0:
            glist = []
            for i in gmags._iterate_signal():
                for j in np.arange(len(i[0])):
                    glist.append(i[0][j])
            gs = np.asarray(glist)
            gsig = Signal1D(gs)
            ghis = gsig.get_histogram(bins=bins)

        else:
            ghis = gmags.get_histogram(bins=bins)

        ghis.axes_manager.signal_axes[0].name = "k"
        ghis.axes_manager.signal_axes[0].units = "$A^{-1}$"

        return ghis

    def cluster(
        self,
        method,
        columns=None,
        column_scale_factors=None,
        min_vectors=None,
        remove_nan=True,
    ):
        """This method clusters a list of vectors both in reciprocal space and in real space.
        The output is a list of vectors with a "label" which defines the cluster that each vector
        belongs to.  Vectors with a label==-1 are outliers which are ignored.

        Parameters
        ----------
        method: sklearn.base.ClusterMixin
            The method used to cluster the vectors
        columns: list
            The columns of the data to use for clustering.
        column_scale_factors: list
            The scale factors to apply to the columns of the data.
        min_vectors: int
            A strict check to limit clusters arising from less than `min_vectors`
            vectors
        remove_nan: bool
            If True, vectors with NaN values are removed before clustering
        """
        if column_scale_factors is None:
            column_scale_factors = [
                1.0,
            ] * self.data.shape[-1]
        if columns is None:
            columns = list(range(self.data.shape[-1]))

        if self.ragged:
            signal_shape = ()
            dtype = object
        else:
            signal_shape = self.axes_manager._signal_shape_in_array
            signal_shape = signal_shape[:-1] + (signal_shape[-1] + 1,)
            dtype = float
        new_signal = self.map(
            cluster,
            inplace=False,
            method=method,
            columns=columns,
            column_scale_factors=column_scale_factors,
            min_vectors=min_vectors,
            ragged=self.ragged,
            remove_nan=remove_nan,
            output_signal_size=signal_shape,
            output_dtype=dtype,
            silence_warnings=True,
        )
        new_signal.column_names = np.append(self.column_names, ["cluster"])
        new_signal.units = np.append(self.units, ["n.a."])

        if not self.has_navigation_axis:
            new_signal.set_signal_type("labeled_diffraction_vectors")

        return new_signal

    @property
    def has_navigation_axis(self):
        return False

    def get_unique_vectors(self, *args, **kwargs):
        """Returns diffraction vectors considered unique by:
        strict comparison, distance comparison with a specified
        threshold, or by clustering using DBSCAN [1].

        Parameters
        ----------
        distance_threshold : float
            The minimum distance between diffraction vectors for them to
            be considered unique diffraction vectors. If
            distance_threshold==0, the unique vectors will be determined
            by strict comparison.
        method : str
            The method to use to determine unique vectors. Valid methods
            are 'strict', 'distance_comparison' and 'DBSCAN'.
            'strict' returns all vectors that are strictly unique and
            corresponds to distance_threshold=0.
            'distance_comparison' checks the distance between vectors to
            determine if some should belong to the same unique vector,
            and if so, the unique vector is iteratively updated to the
            average value.
            'DBSCAN' relies on the DBSCAN [1] clustering algorithm, and
            uses the Eucledian distance metric.
        min_samples : int, optional
            The minimum number of not strictly identical vectors within
            one cluster for the cluster to be considered a core sample,
            i.e. to not be considered noise. Only used for method='DBSCAN'.
        return_clusters : bool, optional
            If True (False is default), the DBSCAN clustering result is
            returned. Only used for method='DBSCAN'.

        References
        ----------
        [1] https://scikit-learn.org/stable/modules/generated/sklearn.cluster.DBSCAN.html

        Returns
        -------
        unique_peaks : DiffractionVectors
            The unique diffraction vectors.
        clusters : DBSCAN
            The results from the clustering, given as class DBSCAN.
            Only returned if method='DBSCAN' and return_clusters=True.
        """
        real_units = self.is_real_units
        flattened_vectors = self.flatten_diffraction_vectors(real_units=real_units)

        return flattened_vectors.get_unique_vectors(*args, **kwargs)

    def filter_magnitude(self, min_magnitude, max_magnitude, *args, **kwargs):
        """
        Filter the diffraction vectors to accept only those with a magnitude within a user specified range.

        Parameters
        ----------
        min_magnitude : float
            Minimum allowed vector magnitude.
        max_magnitude : float
            Maximum allowed vector magnitude.
        *args:
            Arguments to be passed to map().
        **kwargs:
            Keyword arguments to map().

        Returns
        -------
        filtered_vectors : DiffractionVectors
            Diffraction vectors within allowed magnitude tolerances.
        """

        if self.ragged:
            kwargs["output_signal_size"] = ()
            kwargs["output_dtype"] = object

        filtered_vectors = self.map(
            filter_vectors_ragged,
            min_magnitude=min_magnitude,
            max_magnitude=max_magnitude,
            inplace=False,
            silence_warnings=True,
            *args,
            **kwargs,
        )
        return filtered_vectors

    def get_strain_maps(
        self, unstrained_vectors, distance=0.5, columns=None, return_residuals=False
    ):
        """Calculate the strain map from the diffraction vectors.

        Parameters
        ----------
        unstrained_vectors : DiffractionVectors
            The unstrained vector to be used to calculate the strain map.
            If None, the unstrained vector will be calculated from the
            diffraction vectors.
        columns : list
            The columns of the diffraction vectors to be used to calculate
            the strain map. If None, the first two columns will be used.

        Returns
        -------
        strain_map : Signal2D
            The strain map.
        """
        # TODO: remove this import when the function is moved to a utility module

        from pyxem.generators.displacement_gradient_tensor_generator import (
            get_DisplacementGradientMap,
        )

        if isinstance(unstrained_vectors, LazySignal):
            unstrained_vectors.compute()
            unstrained_vectors = unstrained_vectors.data[0]  # remove extra dimension
        if columns is None:
            columns = [0, 1]
        filtered_vectors = self.filter_basis(
            basis=unstrained_vectors, distance=distance, columns=columns
        )
        if isinstance(filtered_vectors, LazySignal):
            filtered_vectors.compute()
        if return_residuals:
            strain_map, residual = get_DisplacementGradientMap(
                filtered_vectors,
                unstrained_vectors,
                return_residuals=return_residuals,
                columns=columns,
            )
            return strain_map.get_strain_maps(), residual
        else:
            strain_map = get_DisplacementGradientMap(
                filtered_vectors,
                unstrained_vectors,
                return_residuals=return_residuals,
                columns=columns,
            )
            return strain_map.get_strain_maps()

    def filter_basis(self, basis, distance=0.5, columns=[0, 1], **kwargs):
        """

        Filter vectors to only the set of vectors which is close to a basis set of vectors.

        If there is no vector within the `distance`
        parameter of the vector np.`nan` will be returned.

        Parameters
        ----------
        basis: array-like or BaseSignal
            The set of vectors to be compared. This should have the same
            number of columns as the length of the ``columns`` parameter.
        distance: float
            The distance between vectors and basis which detemine if the vector
            is associated with the basis vector. If no vector is inside the
            distance np.nan will be returned.
        columns: list
            The columns of the basis to be used for comparison. The default
            is the first two columns (kx, ky) in most cases.
        kwargs: dict
            Any other parameters passed to the `hyperspy.BaseSignal.Map` function.
        Returns
        -------
        vectors: DiffractionVectors or DiffractionVectors2D
            The filtered list of diffraction vectors.  If basis is
            an instance of hyperspy.Signals.BaseSignal and instance of the
            DiffractionVectors class will be returned otherwise an instance
            of the DiffractionVectors2D class will be returned.
        """
        if isinstance(basis, BaseSignal) and (
            basis.axes_manager.navigation_shape == self.axes_manager.navigation_shape
        ):
            kwargs["ragged"] = True
        else:
            if isinstance(basis, LazySignal):
                basis.compute()
                basis = basis.data[0]  # remove extra dimension
            elif isinstance(basis, BaseSignal) and basis.ragged:
                basis = basis.data[0]  # remove extra dimension
            elif isinstance(basis, BaseSignal):
                basis = basis.data
            kwargs["ragged"] = False
        if not isinstance(basis, BaseSignal) and (
            basis.ndim != 2 or basis.shape[0] < 1
        ):
            raise ValueError("The basis must be a 2D array with at least 1 vectors.")

        if not kwargs.get("ragged", False):
            kwargs["output_signal_size"] = (basis.shape[0], self.num_columns)

            kwargs["output_dtype"] = float

        filtered_vectors = self.map(
            filter_vectors_near_basis,
            basis=basis,
            distance=distance,
            inplace=False,
            columns=columns,
            silence_warnings=True,
            **kwargs,
        )
        return filtered_vectors

    def filter_detector_edge(self, exclude_width, *args, **kwargs):
        """Filter the diffraction vectors to accept only those not within a
        user specified proximity to the detector edge.

        Parameters
        ----------
        exclude_width : int
            The width of the region adjacent to the detector edge from which
            vectors will be excluded.
        *args:
            Arguments to be passed to map().
        **kwargs:
            Keyword arguments to map().

        Returns
        -------
        filtered_vectors : DiffractionVectors
            Diffraction vectors within allowed detector region.
        """
        x_threshold = (
            self.scales[0] * (self.detector_shape[0] / 2)
            - self.scales[0] * exclude_width
        )
        y_threshold = (
            self.scales[1] * (self.detector_shape[1] / 2)
            - self.scales[1] * exclude_width
        )
        filtered_vectors = self.map(
            filter_vectors_edge_ragged,
            x_threshold=x_threshold,
            y_threshold=y_threshold,
            inplace=False,
            ragged=True,
            silence_warnings=True,
            *args,
            **kwargs,
        )
        return filtered_vectors

    def to_polar(self, columns=None, **kwargs):
        """Convert the diffraction vectors to polar coordinates.

        Parameters
        ----------
        columns : list
            The columns of the diffraction vectors to be converted to polar
            coordinates. The default is the first two columns (kx, ky) in most
            cases.
        kwargs : dict
            Any other parameters passed to the `hyperspy.signal.BaseSignal.map` function.

        Returns
        -------
        polar_vectors : DiffractionVectors
            Diffraction vectors in polar coordinates.
        """
        polar_vectors = self.map(
            vectors_to_polar,
            inplace=False,
            ragged=self.ragged,
            columns=columns,
            silence_warnings=True,
            **kwargs,
        )
        polar_vectors.set_signal_type("polar_vectors")
        polar_vectors.column_names[0] = "r"
        polar_vectors.column_names[1] = "theta"
        polar_vectors.units[0] = self.units[0]
        polar_vectors.units[1] = "rad"
        return polar_vectors

    def get_diffracting_pixels_map(self, in_range=None, binary=False):
        """Map of the number of vectors at each navigation position.

        Parameters
        ----------
        in_range : tuple
            Tuple (min_magnitude, max_magnitude) the minimum and maximum
            magnitude of vectors to be used to form the map.
        binary : boolean
            If True a binary image with diffracting pixels taking value == 1 is
            returned.

        Returns
        -------
        crystim : Signal2D
            2D map of diffracting pixels.
        """
        if in_range:
            filtered = self.filter_magnitude(in_range[0], in_range[1])
            xim = filtered.map(
                get_npeaks, inplace=False, silence_warnings=True, ragged=False
            ).as_signal2D((0, 1))
        else:
            xim = self.map(
                get_npeaks, silence_warnings=True, inplace=False, ragged=False
            ).as_signal2D((0, 1))
        # Make binary if specified
        if binary is True:
            xim = xim >= 1.0
        # Set properties
        xim = _transfer_navigation_axes_to_signal_axes(xim, self)
        xim.change_dtype("float")
        xim.set_signal_type("signal2d")
        xim.metadata.General.title = "Diffracting Pixels Map"

        return xim

    def to_mask(self, disk_r, signal_axes=None):
        """Convert the diffraction vectors to a N-D mask.

        This can be useful for Orientation Mapping including the fitting of mulitple
        phases.

        Parameters
        ----------
        disk_r : float
            The radius of the disk to be used for the mask.
        signal_axes : list
            The signal axes to be used for the mask. If None, the axes saved
            in the metadata will be used. (if available)

        """
        if signal_axes is None and not self.metadata.has_item("Peaks.signal_axes"):
            raise ValueError("No signal axes provided and no axes saved in metadata.")
        elif signal_axes is None:
            signal_axes = self.metadata.Peaks.signal_axes
            axes = [ax.axis for ax in signal_axes]
        shape = [ax.size for ax in axes]

        def mask_from_vectors(x, disk_r, axes):
            from skimage.draw import disk

            shape = [ax.size for ax in axes]
            mask = np.zeros(shape, dtype=bool)
            vectors = []
            for i, ax in enumerate(axes):
                idx = np.searchsorted(ax, x[:, i], side="left")
                idx[idx > len(ax) - 1] = len(ax) - 1
                idx = idx - (np.abs(x[:, i] - ax[idx - 1]) < np.abs(x[:, i] - ax[idx]))
                vectors.append(idx)
            for v in np.array(
                vectors
            ).T:  # this could maybe be faster if vectorized a bit better
                rr, cc = disk((v[0], v[1]), disk_r, shape=shape)
                mask[rr, cc] = 1
            return mask

        mask = self.map(
            mask_from_vectors,
            inplace=False,
            disk_r=disk_r,
            axes=axes,
            output_signal_size=tuple(shape),
            output_dtype=bool,
            ragged=False,
            silence_warnings=True,
        )
        return mask

    def calculate_cartesian_coordinates(
        self, accelerating_voltage, camera_length, *args, **kwargs
    ):
        """Get cartesian coordinates of the diffraction vectors.

        Parameters
        ----------
        accelerating_voltage : float
            The acceleration voltage with which the data was acquired.
        camera_length : float
            The camera length in meters.
        """
        # Imported here to avoid circular dependency
        from diffsims.utils.sim_utils import get_electron_wavelength

        wavelength = get_electron_wavelength(accelerating_voltage)
        self.cartesian = self.map(
            detector_to_fourier,
            wavelength=wavelength,
            camera_length=camera_length * 1e10,
            inplace=False,
            ragged=True,
            *args,
            **kwargs,
        )


class LazyDiffractionVectors(LazySignal, DiffractionVectors):
    pass



from diffsims.generators.rotation_list_generators import get_beam_directions_grid
from scipy.spatial import ConvexHull
from matplotlib.collections import PolyCollection
from diffsims.generators.simulation_generator import SimulationGenerator
from orix.quaternion import Rotation

import matplotlib.pyplot as plt
import numpy as np
from random import random
from math import cos, sin, pi, radians
from skimage.segmentation import watershed
from copy import deepcopy
from skimage.draw import disk


class CrystalSTEMSimulation:
    """
    A class for creating a 4-D STEM dataset with overlapping nanocrystals.

    Parameters
    ----------
    phase : orix.crystal_map.Phase
        The phase to simulate the diffraction data for.
    num_crystals : int, optional
        The number of nanocrystals to simulate.
    k_range : float, optional
        The range of reciprocal space to simulate.
    crystal_size : int, optional
        The size of the nanocrystals in pixels.
    real_space_pixels : int, optional
        The number of pixels in the real space image to simulate.
    recip_space_pixels : int, optional
        The number of pixels in the reciprocal space image to simulate.
    generator : SimulationGenerator, optional
        The SimulationGenerator to use for calculating the diffraction data.
    max_excitation_error : float, optional
        The maximum excitation error to use for the diffraction data.
    rotations : Rotation, optional
        The rotations to use for the nanocrystals.
    accelerating_voltage : float, optional
        The accelerating voltage to use for the diffraction data.


    Notes
    -----
    This function is not optimized for speed and is intended for generating
    small datasets for testing purposes and for comparing the performance of
    different algorithms.
    """

    def __init__(
        self,
        phase,
        num_crystals=10,
        k_range=1,
        crystal_size=10,
        real_space_pixels=128,
        recip_space_pixels=64,
        generator=None,
        max_excitation_error=0.015,
        rotations=None,
        accelerating_voltage=200,
    ):
        # Set all the parameters
        if generator is None:
            self.generator = SimulationGenerator(
                accelerating_voltage=accelerating_voltage, minimum_intensity=1e-10
            )
        else:
            self.generator = generator
        self.num_crystals = num_crystals
        self.k_range = k_range
        self.crystal_size = crystal_size
        self.real_space_pixels = real_space_pixels
        self.recip_space_pixels = recip_space_pixels
        self.phase = phase

        # Set up the Simulations for calculating the Diffraction Data
        if rotations is None:
            beam_dir = get_beam_directions_grid("cubic", resolution=1)
            r = Rotation.from_euler(beam_dir, degrees=True, direction="crystal2lab")
        else:
            r = rotations

        self.simulation = self.generator.calculate_diffraction2d(
            phase=phase,
            reciprocal_radius=1,
            rotation=r,
            max_excitation_error=max_excitation_error,
        )

        # Randomly set up the center and rotations for the crystals
        self.centers = np.random.randint(
            low=crystal_size + 5,
            high=real_space_pixels - (crystal_size + 5),
            size=(num_crystals, 2),
        )
        self.random_rotations = np.random.randint(0, len(r.data), num_crystals)
        self.random_inplane = np.random.random(num_crystals) * np.pi

        # Get the Nano crystal Vectors

        (
            self.vectors,
            self.coordinates,
            self.real_space_vectors,
            self.intensities,
        ) = self.get_nano_crystal_vectors()

    def get_nano_crystal_vectors(self):
        """
        Get the nano crystal vectors

        """
        real_space_pixels = self.real_space_pixels
        coordinates = []
        real_space_pos = []
        vectors = []
        intens = []
        # For each crystal randomly rotate a
        for rot, center, inplane in zip(
            self.random_rotations, self.centers, self.random_inplane
        ):
            # Getting the simulation
            cor = self.simulation.irot[rot].coordinates.data[:, :2]
            inten = self.simulation.irot[rot].coordinates.intensity

            vector_pos = np.argwhere(
                create_blob(
                    center=center,
                    size=self.crystal_size,
                    real_space_pixels=real_space_pixels,
                )
            )
            real_space_pos.append(vector_pos)

            # Randomly rotating in plane
            corx = cor[:, 0] * np.cos(inplane) - cor[:, 1] * np.sin(inplane)
            cory = cor[:, 0] * np.sin(inplane) + cor[:, 1] * np.cos(inplane)
            cors = np.stack([corx, cory, inten], axis=1)
            coordinates.append(cors)
            intens.append(inten)
            # Create a list of 4-D Vectors describing the extent of the
            v = [list(v) + list(c) for v in vector_pos for c in cors]
            vectors.append(v)
            intens.append(intens)
        # Unpack the vectors for plotting
        vectors = np.array([v for vector in vectors for v in vector])
        return vectors, coordinates, real_space_pos, intens

    def get_coords(self, low=0.0, high=0.9):
        filtered_coords = []
        for co in self.coordinates:
            norms = np.linalg.norm(co[:, :2], axis=1)
            within_range = (norms > low) * (norms < high)
            filtered_coords.append(co[within_range])
        return filtered_coords

    def make_4d_stem(
        self,
        num_electrons=1,
        electron_gain=1,
        noise_level=0.1,
        radius=5,
    ):
        """
        Make a 4-D Dataset

        Parameters
        ----------
        num_electrons : int, optional
            The number of electrons per pixel on average
        electron_gain : float, optional
            The gain of the detector. Each electron will be multiplied by this value
        noise_level : float, optional
            The level of noise to add to the dataset. Noise will fall between 0 and this value
        radius : int, optional
            The radius of the disks to add to the dataset

        Returns
        -------
        np.ndarray
            A 4-D array representing the 4-D STEM dataset
        """
        real_space_pixels = self.real_space_pixels
        recip_space_pixels = self.recip_space_pixels
        k_range = self.k_range

        vectors_by_index = convert_to_by_index(
            self.vectors, real_space=self.real_space_pixels
        )

        arr = np.zeros(
            (
                real_space_pixels,
                real_space_pixels,
                recip_space_pixels,
                recip_space_pixels,
            )
        )

        scale = (recip_space_pixels / 2) / k_range
        center = recip_space_pixels / 2
        for i in np.ndindex((real_space_pixels, real_space_pixels)):
            vector = vectors_by_index[i]
            im = np.zeros((self.recip_space_pixels, self.recip_space_pixels))
            for v in vector:
                im = add_disk(
                    im, center=v[:2] * scale + center, radius=radius, intensity=v[2]
                )
            arr[i] = im

        arr = arr * num_electrons
        arr = np.random.poisson(arr) * electron_gain

        noise = np.random.random(arr.shape) * noise_level
        arr = arr + noise
        return arr

    def plot_real_space(
        self,
        ax=None,
        remove_below_n=None,
        remove_non_symmetric=False,
        **kwargs,
    ):
        """
        Plot the real space image of the nanocrystals. This is a helpful ground truth


        Parameters
        ----------
        ax: matplotlib.axes.Axes, optional
            The axes to plot the image on
        remove_below_n: int
            Remove nanocrystals with less than n k-vectors. This is helpful for removing nanocrystals
            that don't have many characteristic k-vectors
        remove_non_symmetric: bool
            Remove nanocrystals that have non-symmetric k-vectors. This is helpful for removing nanocrystals
            that don't have symmetric diffraction
        kwargs:
            Additional keyword arguments to pass to the `matplotlib.collections.PolyCollection`
            class.

        Returns
        -------

        """
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        colors = [
            "blue",
            "green",
            "red",
            "yellow",
            "purple",
            "orange",
            "violet",
            "indigo",
            "black",
            "grey",
        ]
        verts = []
        if remove_below_n is not None:
            vectors = [
                v
                for v, c in zip(self.real_space_vectors, self.coordinates)
                if len(c) > remove_below_n
            ]
        else:
            vectors = self.real_space_vectors

        if remove_non_symmetric:
            new_vectors = []
            for c, v in zip(self.coordinates, self.real_space_vectors):
                un, counts = np.unique(
                    np.round(np.linalg.norm(c[:, :2], axis=1), 1), return_counts=True
                )
                if np.any(counts > 2):
                    new_vectors.append(v)
            vectors = new_vectors

        for v in vectors:
            hull = ConvexHull(v)
            vert = hull.points[hull.vertices]
            verts.append(vert)
        p = PolyCollection(verts, color=colors, **kwargs)
        ax.add_collection(p)
        ax.set_xlim(0, self.real_space_pixels)
        ax.set_ylim(0, self.real_space_pixels)
        ax.set_xticks([])
        ax.set_yticks([])
        return ax

    def plot_example_dp(
        self,
        rotation_ind=0,
        num_electrons=1,
        electron_gain=1,
        noise_level=0.1,
        pixels=64,
        reciprocal_radius=1,
        ax=None,
        threshold=0.8,
        disk_r=5,
        **kwargs,
    ):
        """
        Plot an example diffraction pattern

        Parameters
        ----------
        rotation_ind:
            The index of the rotation to plot
        num_electrons:
            The number of electrons per pixel on average
        electron_gain:
            The gain of the detector. Each electron will be multiplied by this value
        noise_level:
            The level of noise to add to the dataset. Noise will fall between 0 and this value
        pixels:
            The number of pixels in the diffraction pattern
        reciprocal_radius:
            The reciprocal radius to plot
        ax: matplotlib.axes.Axes, optional
            The axes to plot the image on
        threshold:
            The threshold to use for the image
        disk_r: int
            The radius of the disks to add to the dataset
        kwargs:
            Additional keyword arguments to pass to the `matplotlib.pyplot.imshow`
            function.
        """
        if ax is None:
            fig, ax = plt.subplots(1, 1)

        sim = self.simulation.irot[rotation_ind]
        img = np.zeros((pixels, pixels))
        coors = sim.coordinates.data[:, :2]
        intens = sim.coordinates.intensity
        scale = (pixels / 2) / reciprocal_radius
        center = pixels / 2

        for c, inten in zip(coors, intens):
            img = add_disk(
                img, center=c * scale + center, radius=disk_r, intensity=inten
            )
        img = img * num_electrons

        img = np.random.poisson(img) * electron_gain
        img = img + np.random.random(img.shape) * noise_level
        if threshold is not None:
            kwargs["vmax"] = np.max(img) * threshold
        ax.imshow(
            img,
            extent=(
                -reciprocal_radius,
                reciprocal_radius,
                -reciprocal_radius,
                reciprocal_radius,
            ),
            **kwargs,
        )
        return ax


def create_blob(center, size, real_space_pixels):
    """
    A utility function to create a blob in the real space image

    Parameters
    ----------
    center:
        The center of the blob
    size: int
        The "radius" of the blob in pixels
    real_space_pixels:
        The number of pixels in the real space image to account for
    """
    N = 4
    amps = [random() * (1 / (2 * N)) for _ in range(N)]
    phases = [random() * 2 * pi for _ in range(N)]

    points = np.empty((360, 2))
    img = np.zeros((real_space_pixels, real_space_pixels))
    for deg in range(360):
        alpha = radians(deg)
        radius = 1 + sum([amps[i] * cos((i + 1) * alpha + phases[i]) for i in range(N)])
        points[deg, 0] = (cos(alpha) * radius * size) + center[0]
        points[deg, 1] = (sin(alpha) * radius * size) + center[1]
    points = np.array(
        np.round(
            points,
        ),
        dtype=int,
    )
    img[points[:, 0], points[:, 1]] = 1
    img = watershed(img) == 2
    return img


def convert_to_by_index(peaks, real_space):
    """
    Convert  a vector of vectors to a indexable array of vectors in real space for
    creating a 4-D STEM dataset.

    Parameters
    ----------
    peaks: np.ndarray
        The array of vectors
    real_space:
        The number of pixels in the real space image to account for
    """
    new_peaks = deepcopy(peaks)
    ind = np.lexsort((new_peaks[:, 1], new_peaks[:, 0]))
    sorted_peaks = new_peaks[ind]
    by_ind_peaks = np.empty((real_space, real_space), dtype=object)
    low_x_ind = np.searchsorted(sorted_peaks[:, 0], range(0, real_space), side="left")
    high_x_ind = np.searchsorted(
        sorted_peaks[:, 0], range(1, real_space + 1), side="left"
    )
    for i, (lo_x, hi_x) in enumerate(zip(low_x_ind, high_x_ind)):
        x_inds = sorted_peaks[lo_x:hi_x]
        low_y_ind = np.searchsorted(x_inds[:, 1], range(0, real_space), side="left")
        high_y_ind = np.searchsorted(
            x_inds[:, 1], range(1, real_space + 1), side="left"
        )
        for j, (lo_y, hi_y) in enumerate(zip(low_y_ind, high_y_ind)):
            by_ind_peaks[i, j] = x_inds[lo_y:hi_y, 2:]
    return by_ind_peaks


def add_disk(image, center, radius, intensity):
    """
    Add a disk to an image

    Parameters
    ----------
    image: np.ndarray
        The image to add the disk to
    center:
        The center of the disk
    radius:
        The radius of the disk
    intensity:
        The intensity of the disk

    """
    disk_image = np.zeros_like(image)
    rr, cc = disk(center=center, radius=radius, shape=image.shape)
    disk_image[rr, cc] = intensity  # expected 1 electron per pixel
    image = disk_image + image
    return image



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np

from pyxem.signals import ElectronDiffraction1D, ReducedIntensity1D
from pyxem.generators import ReducedIntensityGenerator1D


@pytest.fixture
def red_int_generator():
    data = np.arange(10, 0, -1).reshape(1, 10) * np.arange(1, 5).reshape(4, 1)
    data = data.reshape(2, 2, 10)
    rp = ElectronDiffraction1D(data)
    rigen = ReducedIntensityGenerator1D(rp)
    return rigen


def test_init(red_int_generator):
    assert isinstance(red_int_generator, ReducedIntensityGenerator1D)


def test_scattering_calibration(red_int_generator):
    calib = 0.1
    red_int_generator.set_diffraction_calibration(calibration=calib)
    s_scale = red_int_generator.signal.axes_manager.signal_axes[0].scale
    assert s_scale == calib
    offset = 0.05
    red_int_generator.set_diffraction_offset(offset=offset)
    s_offset = red_int_generator.signal.axes_manager.signal_axes[0].offset
    assert s_offset == offset


def test_fit_atomic_scattering(red_int_generator):
    calib = 0.1
    red_int_generator.set_diffraction_calibration(calibration=calib)
    elements = ["Cu"]
    fracs = [1]
    assert red_int_generator.background_fit is None
    assert red_int_generator.normalisation is None
    red_int_generator.fit_atomic_scattering(elements=elements, fracs=fracs)

    assert red_int_generator.background_fit.data.shape == (2, 2, 10)
    assert red_int_generator.normalisation.data.shape == (2, 2, 10)


def test_set_cutoff(red_int_generator):
    s_min, s_max = 0, 8
    red_int_generator.set_s_cutoff(s_min, s_max)
    assert red_int_generator.cutoff == [s_min, s_max]


def test_subtract_bkgd(red_int_generator):
    bkgd_pattern = np.arange(10, 0, -1).reshape(10)
    red_int_generator.subtract_bkgd_pattern(bkgd_pattern)

    expected = np.arange(10, 0, -1).reshape(1, 10) * np.arange(0, 4).reshape(4, 1)
    expected = expected.reshape(2, 2, 10)

    assert np.array_equal(red_int_generator.signal.data, expected)


def test_mask_from_bkgd(red_int_generator):
    mask_pattern = np.arange(10, 0, -1).reshape(10)
    mask_threshold = 6.5
    red_int_generator.mask_from_bkgd_pattern(
        mask_pattern, mask_threshold=mask_threshold
    )

    expected = np.arange(10, 0, -1).reshape(1, 10) * np.arange(1, 5).reshape(4, 1)
    expected = expected.reshape(2, 2, 10)
    expected[:, :, :4] = 0

    assert np.array_equal(red_int_generator.signal.data, expected)


def test_mask_reduced_intensity(red_int_generator):
    mask_pattern = np.ones(10)
    mask_pattern[:4] = 0
    red_int_generator.mask_reduced_intensity(mask_pattern)

    expected = np.arange(10, 0, -1).reshape(1, 10) * np.arange(1, 5).reshape(4, 1)
    expected = expected.reshape(2, 2, 10)
    expected[:, :, :4] = 0

    assert np.array_equal(red_int_generator.signal.data, expected)


def test_incorrect_mask(red_int_generator):
    mask_pattern = np.ones(10)
    mask_pattern[:4] = 0
    mask_pattern[8] = 2
    with pytest.raises(
        ValueError, match="Masking array does not consist of zeroes and ones"
    ):
        red_int_generator.mask_reduced_intensity(mask_pattern)


def test_get_reduced_intensity(red_int_generator):
    red_int_generator.set_diffraction_calibration(calibration=0.1)
    red_int_generator.fit_atomic_scattering(elements=["Cu"], fracs=[1])
    ri = red_int_generator.get_reduced_intensity()
    assert isinstance(ri, ReducedIntensity1D)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
import scipy.constants
from scipy.ndimage import rotate
from hyperspy._signals.lazy import LazySignal
from hyperspy.axes import UniformDataAxis

import pyxem.utils._beam_shift_tools as bst
from pyxem.utils.plotting import make_color_wheel_marker
from pyxem.signals import DiffractionVectors1D
from pyxem.utils._deprecated import deprecated
from diffsims.utils.sim_utils import get_electron_wavelength
from scipy.constants import c, e, m_e


class BeamShift(DiffractionVectors1D):
    """Signal class for working with shift of the direct beam."""

    _signal_type = "beam_shift"

    @deprecated(
        since="0.19.0",
        alternative="get_linear_plane",
        alternative_is_function=True,
        removal="1.0.0",
    )
    def make_linear_plane(self, **kwargs):
        s_linear_plane = self.get_linear_plane(**kwargs)
        self.data = s_linear_plane.data
        self.events.data_changed.trigger(None)

    def get_linear_plane(
        self,
        mask=None,
        fit_corners=None,
        initial_values=None,
        constrain_magnitude_variance=False,
    ):
        """Fit linear planes to the beam shifts, and returns a BeamShift signal
        with the planes.

        In many scanning transmission electron microscopes, the center position of the
        diffraction pattern will change as a function of the scan position. This is most
        apparent when scanning over large regions (100+ nanometers). Thus, when working
        with datasets, it is typically necessary to correct for this.
        However, other effects can also affect the apparent center point, like
        diffraction contrast. So while it is possible to correct for the beam shift
        by finding the center position in each diffraction pattern, this can lead to
        features such as interfaces or grain boundaries affecting the centering of the
        diffraction pattern. As the shift caused by the scan system is slow and
        monotonic, it can often be approximated by fitting linear planes to
        the x- and y- beam shifts.

        In addition, for regions within the scan where the center point of the direct
        beam is hard to ascertain precisely, for example in very thick or heavily
        diffracting regions, a mask can be used to ignore fitting the plane to
        these regions.

        This method does this, and returns a new BeamShift signal with these
        fitted planes. The beam shift signal can then be directly used in the
        :meth:`~pyxem.signals.Diffraction2D.center_direct_beam` method.

        Note that for very large regions, this linear plane will probably not
        approximate the beam shift very well. In those cases a higher order plane
        will likely be necessary. Alternatively, a vacuum scan with exactly the same
        scanning parameters should be used.

        Parameters
        ----------
        mask : HyperSpy signal, optional
            Must be the same shape as the navigation dimensions of the beam
            shift signal. The True values will be masked.
        fit_corners : float, optional
            Make a mask so that the planes are fitted to the corners of the
            signal. This mush be set with a number, like 0.05 (5%) or 0.10 (10%).
        initial_values : array of floats, optional
            Initial guess for the plane parameters. Useful to vary if the plane fitting
            does not give desirable results.
            The horizontal- and vertical-shifts are described by two linear planes
            with three parameters. The two first parameters, d/dx and d/dy, are the changes
            in horizontal-shift as you move one position in the navigation space in
            respectively the x- and y-directions. I. e. they are the steps in
            horizontal-shift as you change x- or y-coordinates. The third parameter,
            shift_0, is the horizontal-shift in the (0, 0) navigation position. The
            vertical-shift are described by similar parameters. In this argument
            supply the plane parameters in the following way, with the first three
            being for the horizontal-shift and the rest for the vertical-shift:
            [d/dx, d/dy, shift_0, d/dx, d/dy, shift_0]
            Currently only implemented for the case when `constrain_magnitude_variance`
            is `True`.
        constrain_magnitude_variance : bool, optional
            Fits the linear planes to deflections with constant magnitude. By default
            set to `False`.
            In the presence of electromagnetic fields in the sample area, least squares
            fitting can give inaccurate results. If the region is expected to have
            uniform field strength, we can fit planes by trying to minimise the variance
            of the magnitudes, giving a constant deflection magnitude.
            Note that for this to work several field directions must be present. Extra
            care must be taken in presence of significant noise, such as with a mask.
            If desirable results are not found, try varying the `initial_values`
            parameter.

        Examples
        --------
        >>> s = pxm.signals.BeamShift(np.random.randint(0, 99, (100, 120, 2)))
        >>> s_mask = hs.signals.Signal2D(np.zeros((100, 120), dtype=bool))
        >>> s_mask.data[20:-20, 20:-20] = True
        >>> s_linear_plane = s.get_linear_plane(mask=s_mask)

        For magnetic DPC signals, one would typically want to subtract these linear
        planes from the original signal

        >>> s_corr = s - s_linear_plane

        """
        if self._lazy:
            raise ValueError(
                "get_linear_plane is not implemented for lazy signals, "
                "run compute() first"
            )
        if self.axes_manager.navigation_dimension != 2:
            raise NotImplementedError(
                "get_linear_plane is only implemented for signals with "
                "2 navigation dimensions, not {0}".format(
                    self.axes_manager.navigation_dimension
                )
            )

        if (mask is not None) and (fit_corners is not None):
            raise ValueError("Only mask or fit_to_corners can be set.")
        if fit_corners is not None:
            mask = bst._get_corner_mask(self.isig[0], corner_size=fit_corners)

        s_shift_x = self.isig[0].T
        s_shift_y = self.isig[1].T
        if mask is not None:
            mask = mask.__array__()
            if mask.dtype != bool:
                raise ValueError("mask needs to have a datatype of bool")
        if constrain_magnitude_variance:
            plane_image = bst._get_linear_plane_by_minimizing_magnitude_variance(
                self, mask=mask, initial_values=initial_values
            )
        else:
            plane_image_x = bst._get_linear_plane_from_signal2d(s_shift_x, mask=mask)
            plane_image_y = bst._get_linear_plane_from_signal2d(s_shift_y, mask=mask)
            plane_image = np.stack((plane_image_x, plane_image_y), -1)
        s_bs = self._deepcopy_with_new_data(plane_image)
        return s_bs

    def get_bivariate_histogram(
        self, histogram_range=None, masked=None, bins=200, spatial_std=3
    ):
        """
        Useful for finding the distribution of the beam shifts.
        Especially useful for magnetic signals, as it can tell us about
        the domain distribution.

        Parameters
        ----------
        histogram_range : tuple, optional
            Set the minimum and maximum of the histogram range.
            Default is setting it automatically.
        masked : 2-D NumPy bool array, optional
            Mask parts of the data. The array must be the same
            size as the signal. The True values are masked.
            Default is not masking anything.
        bins : integer, default 200
            Number of bins in the histogram
        spatial_std : number, optional
            If histogram_range is not given, this value will be
            used to set the automatic histogram range.
            Default value is 3.

        Returns
        -------
        s_hist : HyperSpy Signal2D

        Examples
        --------
        >>> s = pxm.data.dummy_data.get_stripe_pattern_beam_shift_signal()
        >>> s_hist = s.get_bivariate_histogram()
        >>> s_hist.plot()

        """
        x_position = self.isig[0].data
        y_position = self.isig[1].data
        s_hist = bst._make_bivariate_histogram(
            x_position,
            y_position,
            histogram_range=histogram_range,
            masked=masked,
            bins=bins,
            spatial_std=spatial_std,
        )
        s_hist.metadata.General.title = "Bivariate histogram of {0}".format(
            self.metadata.General.title
        )
        return s_hist

    @property
    def beam_energy(self):
        try:
            return self.metadata.get_item("Acquisition_instrument.TEM.beam_energy")
        except KeyError:
            return None

    @beam_energy.setter
    def beam_energy(self, beam_energy):
        self.metadata.set_item("Acquisition_instrument.TEM.beam_energy", beam_energy)
        self.metadata.set_item(
            "Acquisition_instrument.TEM.wavelength",
            get_electron_wavelength(beam_energy),
        )

    def calibrate_electric_shifts(self, thickness):
        """
        Calibrate the shifts into the electric field

        m_0 * gamma * |v|^2 * theta / e*t = E

        where m_0 is the electron rest mass, gamma is the Relativistic factor, t is the
        thickness m_0 is the electron rest mass, e is the electron charge, and theta
        is the deflection angle in radians.

        Parameters
        ----------
        thickness : float
            The thickness of the sample in nm.
        """
        if self.units != ["mrad", "mrad"]:
            raise ValueError(
                "The beam shifts must be in mrad units to calibrate to electric field."
                "Please call `pixels_to_calibrated_units` first."
            )
        if self.beam_energy is None:
            raise ValueError(
                "The beam energy must be set in the metadata to calibrate to electric field."
                "Please set the beam energy using `beam_energy` property."
            )
        thickness = thickness * 1e-9  # Convert thickness from nm to m
        volts = self.beam_energy * 1000
        velocity = c * np.sqrt(1 - (1 / (1 + (volts * e / (m_e * c**2))) ** 2))  # m/s
        y = 1 / np.sqrt(1 - (velocity**2 / c**2))  # Relativistic factor
        shifts = ((self / 1000) * m_e * y * np.abs(velocity) ** 2) / (
            -e * thickness
        )  # convert to rad
        shifts = shifts / 1e8  # Convert to MV/cm
        shifts.units = ["MV/cm", "MV/cm"]
        shifts.metadata.Signal.quantity = "MV/cm"
        return shifts

    def pixels_to_calibrated_units(self, signal_axes=None, inplace=False, **kwargs):
        """Convert the beam shifts from pixels to calibrated units using the
        signal axes passed or saved in the metadata.

        Parameters
        ----------
        signal_axes : list of UniformDataAxis, optional
            The signal axes to use for the conversion. If not provided, the
            signal axes saved in the metadata will be used.
        inplace : bool, optional
            If True, the data will be replaced with the converted data.
            Default is False.
        **kwargs : dict
            Additional keyword arguments to pass to the map method.

        """
        if signal_axes is None:
            try:
                signal_axes = self.metadata.Shifts.signal_axes
            except AttributeError:
                raise ValueError(
                    "No signal axes found in metadata. Please provide signal_axes."
                )
        signal_axes = [
            s.convert_to_uniform_axis() if not isinstance(s, UniformDataAxis) else s
            for s in signal_axes
        ]
        scales = [s.scale for s in signal_axes]

        cal_com = self.map(
            lambda x, scale: x * scale,
            output_dtype=float,
            output_signal_size=(2,),
            scale=np.array(scales),
            inplace=inplace,
            **kwargs
        )
        cal_com.units = [s.units for s in signal_axes]
        return cal_com

    def get_magnitude_signal(
        self, autolim=True, autolim_sigma=4, magnitude_limits=None
    ):
        """Get beam shift magnitude image visualized as greyscale.

        Converts the x and y beam shifts into a magnitude map, showing the
        magnitude of the beam shifts.

        Useful for visualizing magnetic domain structures.

        Parameters
        ----------
        autolim : bool, default True
        autolim_sigma : float, default 4
        magnitude_limits : tuple of floats, default None
            Manually sets the value limits for the magnitude signal.
            For this, autolim needs to be False.

        Returns
        -------
        magnitude_signal : HyperSpy 2D signal

        Examples
        --------
        >>> s = pxm.data.dummy_data.get_simple_beam_shift_signal()
        >>> s_magnitude = s.get_magnitude_signal()
        >>> s_magnitude.plot()

        See Also
        --------
        get_magnitude_phase_signal : Signal showing both phase and magnitude
        get_phase_signal : Signal showing the phase

        """

        def magnitude_calc_with_map(image):
            x, y = image
            mag = np.hypot(x, y)
            return mag

        s_magnitude = self.map(magnitude_calc_with_map, inplace=False)
        s_magnitude = s_magnitude.T

        if autolim:
            if magnitude_limits is not None:
                raise ValueError(
                    "If autolim==True then `magnitude_limits` must be set to None"
                )

            magnitude_limits = bst._get_limits_from_array(
                s_magnitude.data, sigma=autolim_sigma
            )
        if magnitude_limits is not None:
            np.clip(
                s_magnitude.data,
                magnitude_limits[0],
                magnitude_limits[1],
                out=s_magnitude.data,
            )

        s_magnitude.metadata.General.title = "Magnitude of {0}".format(
            self.metadata.General.title
        )
        return s_magnitude

    def phase_retrieval(self, method="kottler", mirroring=False, mirror_flip=False):
        """Retrieve the phase from two orthogonal phase gradients.

        Parameters
        ----------
        method : 'kottler', 'arnison' or 'frankot', optional
            the formula to use, kottler [1]_ , arnison [2]_ and frankot [3]_
            are available. The default is 'kottler'.
        mirroring : bool, optional
            whether to mirror the phase gradients before Fourier transformed.
            Attempt to reduce boundary effect. The default is False.
        mirror_flip : bool, optional
            only active when 'mirroring' is True. Flip the direction of the
            derivatives which results in negation during signal mirroring.
            The default is False. If the retrieved phase is not sensible after
            mirroring, set this to True may resolve it.

        Raises
        ------
        ValueError
            If the method is not implemented

        Returns
        -------
        signal : HyperSpy 2D signal
            the phase retrieved.

        References
        ----------
        .. [1] Kottler, C., David, C., Pfeiffer, F. and Bunk, O., 2007. A
           two-directional approach for grating based differential phase contrast
           imaging using hard x-rays. Optics Express, 15(3), p.1175. (Equation 4)

        .. [2] Arnison, M., Larkin, K., Sheppard, C., Smith, N. and
           Cogswell, C., 2004. Linear phase imaging using differential
           interference contrast microscopy. Journal of Microscopy, 214(1),
           pp.7-12. (Equation 6)

        .. [3] Frankot, R. and Chellappa, R., 1988. A method for enforcing
           integrability in shape from shading algorithms.
           IEEE Transactions on Pattern Analysis and Machine Intelligence,
           10(4), pp.439-451. (Equation 21)

        Examples
        --------
        >>> s = pxm.data.dummy_data.get_magnetic_square_beam_shift_signal()
        >>> s_phase = s.phase_retrieval()
        >>> s_phase.plot()

        """

        method = method.lower()
        if method not in ("kottler", "arnison", "frankot"):
            raise ValueError(
                "Method '{}' not recognised. 'kottler', 'arnison'"
                " and 'frankot' are available.".format(method)
            )

        # get x and y phase gradient
        dx = self.isig[0].data
        dy = self.isig[1].data

        # attempt to reduce boundary effect
        if mirroring:
            Ax = dx
            Bx = np.flip(dx, axis=1)
            Cx = np.flip(dx, axis=0)
            Dx = np.flip(dx)

            Ay = dy
            By = np.flip(dy, axis=1)
            Cy = np.flip(dy, axis=0)
            Dy = np.flip(dy)

            # the -ve depends on the direction of derivatives
            if not mirror_flip:
                dx = np.bmat([[Ax, -Bx], [Cx, -Dx]]).A
                dy = np.bmat([[Ay, By], [-Cy, -Dy]]).A
            else:
                dx = np.bmat([[Ax, Bx], [-Cx, -Dx]]).A
                dy = np.bmat([[Ay, -By], [Cy, -Dy]]).A

        nc, nr = dx.shape[1], dx.shape[0]

        # get scan step size
        calX = np.diff(self.axes_manager.navigation_axes[0].axis).mean()
        calY = np.diff(self.axes_manager.navigation_axes[1].axis).mean()

        # construct Fourier-space grids
        kx = (2 * np.pi) * np.fft.fftshift(np.fft.fftfreq(nc))
        ky = (2 * np.pi) * np.fft.fftshift(np.fft.fftfreq(nr))
        kx_grid, ky_grid = np.meshgrid(kx, ky)

        if method == "kottler":
            gxy = dx + 1j * dy
            numerator = np.fft.fftshift(np.fft.fft2(gxy))
            denominator = 2 * np.pi * 1j * (kx_grid + 1j * ky_grid)
        elif method == "arnison":
            gxy = dx + 1j * dy
            numerator = np.fft.fftshift(np.fft.fft2(gxy))
            denominator = 2j * (
                np.sin(2 * np.pi * calX * kx_grid)
                + 1j * np.sin(2 * np.pi * calY * ky_grid)
            )
        elif method == "frankot":
            kx_grid /= calX
            ky_grid /= calY
            fx = np.fft.fftshift(np.fft.fft2(dx))
            fy = np.fft.fftshift(np.fft.fft2(dy))
            # weights in x,y directins, currently hardcoded, but easy to extend if required
            wx, wy = 0.5, 0.5

            numerator = -1j * (wx * kx_grid * fx + wy * ky_grid * fy)
            denominator = wx * kx_grid**2 + wy * ky_grid**2

        # handle the division by zero in the central pixel
        # set the undefined/infinity pixel to 0
        with np.errstate(divide="ignore", invalid="ignore"):
            res = numerator / denominator
        res = np.nan_to_num(res, nan=0, posinf=0, neginf=0)

        retrieved = np.fft.ifft2(np.fft.ifftshift(res)).real

        # get 1/4 of the result if mirroring
        if mirroring:
            M, N = retrieved.shape
            retrieved = retrieved[: M // 2, : N // 2]

        signal = self._deepcopy_with_new_data(retrieved)
        signal._remove_axis(-1)
        signal = signal.T
        signal.metadata.General.title = "Phase retrieval of {0}".format(
            self.metadata.General.title
        )
        return signal

    def get_phase_signal(self, rotation=None, add_color_wheel_marker=True):
        """Get beam shift phase image visualized using continuous color scale.

        Converts the x and y beam shifts into an RGB array, showing the
        direction of the beam shifts.

        Useful for visualizing magnetic domain structures.

        Parameters
        ----------
        rotation : float, optional
            In degrees. Useful for correcting the mismatch between
            scan direction and diffraction pattern rotation.
        autolim : bool, default True
        autolim_sigma : float, default 4

        Returns
        -------
        phase_signal : HyperSpy 2D RGB signal

        Examples
        --------
        >>> s = pxm.data.dummy_data.get_simple_beam_shift_signal()
        >>> s_color = s.get_phase_signal(rotation=20)
        >>> s_color.plot()

        See Also
        --------
        get_magnitude_phase_signal : Signal showing both phase and magnitude
        get_magnitude_signal : Signal showing the magnitude

        """

        if self.axes_manager.navigation_dimension != 2:
            raise ValueError("get_phase_signal only works with 2 navigation dimensions")
        # Rotate the phase by -30 degrees in the color "wheel", to get better
        # visualization in the vertical and horizontal direction.
        if rotation is None:
            rotation = -30
        else:
            rotation = rotation - 30

        phase = np.arctan2(self.isig[0].data, self.isig[1].data) % (2 * np.pi)
        rgb_array = bst._get_rgb_phase_array(phase=phase, rotation=rotation)
        rgb_array_16bit = rgb_array * (2**16 - 1)
        s_rgb = self._deepcopy_with_new_data(rgb_array_16bit)
        s_rgb.change_dtype("uint16")
        s_rgb.change_dtype("rgb16")
        if add_color_wheel_marker:
            # Add a color wheel marker to the signal
            color_wheel_marker = make_color_wheel_marker(
                rotation=rotation, only_phase=True
            )
            s_rgb.add_marker(color_wheel_marker, permanent=True)
        return s_rgb

    @deprecated(
        since="0.19.0",
        alternative="get_magnitude_phase_signal",
        alternative_is_function=True,
        removal="1.0.0",
    )
    def get_color_signal(self, **kwargs):
        return self.get_magnitude_phase_signal(**kwargs)

    def get_magnitude_phase_signal(
        self,
        rotation=None,
        autolim=True,
        autolim_sigma=4,
        magnitude_limits=None,
        add_color_wheel_marker=True,
    ):
        """Get beam shift image visualized using continuous color scale.

        Converts the x and y beam shifts into an RGB array, showing the
        magnitude and phase (direction) of the beam shifts.

        Useful for visualizing magnetic domain structures.

        Parameters
        ----------
        rotation : float, optional
            In degrees. Useful for correcting the mismatch between
            scan direction and diffraction pattern rotation.
        autolim : bool, default True
        autolim_sigma : float, default 4
        magnitude_limits : tuple of floats, default None
            Manually sets the value limits for the color signal.
            For this, autolim needs to be False.

        Returns
        -------
        color_signal : HyperSpy 2D RGB signal

        Examples
        --------
        >>> s = pxm.data.dummy_data.get_simple_beam_shift_signal()
        >>> s_magnitude_phase = s.get_magnitude_phase_signal()
        >>> s_magnitude_phase.plot()

        Rotate the beam shift by 30 degrees

        >>> s_magnitude_phase = s.get_magnitude_phase_signal(rotation=30)

        See Also
        --------
        get_magnitude_signal : Signal showing the magnitude
        get_phase_signal : Signal showing the phase

        """
        # Rotate the phase by -30 degrees in the color "wheel", to get better
        # visualization in the vertical and horizontal direction.
        if rotation is None:
            rotation = -30
        else:
            rotation = rotation - 30
        phase = np.arctan2(self.isig[0].data, self.isig[1].data) % (2 * np.pi)
        magnitude = np.hypot(self.isig[0].data, self.isig[1].data)

        if autolim:
            if magnitude_limits is not None:
                raise ValueError(
                    "If autolim==True then `magnitude_limits` must be set to None"
                )

            magnitude_limits = bst._get_limits_from_array(
                magnitude, sigma=autolim_sigma
            )
        rgb_array = bst._get_rgb_phase_magnitude_array(
            phase=phase,
            magnitude=magnitude,
            rotation=rotation,
            magnitude_limits=magnitude_limits,
        )
        rgb_array_16bit = rgb_array * (2**16 - 1)
        s_rgb = self._deepcopy_with_new_data(rgb_array_16bit)
        s_rgb.change_dtype("uint16")
        s_rgb.change_dtype("rgb16")

        if add_color_wheel_marker:
            # Add a color wheel marker to the signal
            color_wheel_marker = make_color_wheel_marker(rotation=rotation)
            s_rgb.add_marker(color_wheel_marker, permanent=True, plot_marker=False)
        return s_rgb

    def rotate_beam_shifts(self, angle):
        """Rotate the beam shift vector.

        Parameters
        ----------
        angle : float
            Clockwise rotation in degrees

        Returns
        -------
        shift_rotated_signal : BeamShift

        Example
        -------

        Rotate beam shifts by 10 degrees clockwise

        >>> s = pxm.data.dummy_data.get_simple_beam_shift_signal()
        >>> s_new = s.rotate_beam_shifts(10)
        >>> s_new.plot()

        """
        angle_rad = np.deg2rad(angle)
        x, y = self.isig[0].data, self.isig[1].data
        x_new = x * np.cos(angle_rad) - y * np.sin(angle_rad)
        y_new = x * np.sin(angle_rad) + y * np.cos(angle_rad)
        s_new = self._deepcopy_with_new_data(np.stack((x_new, y_new), axis=-1))
        return s_new

    def rotate_scan_dimensions(self, angle, reshape=False):
        """Rotate the scan dimensions by angle.

        Parameters
        ----------
        angle : float
            Clockwise rotation in degrees

        Returns
        -------
        rotated_signal : BeamShift

        Example
        -------

        Rotate data by 10 degrees clockwise

        >>> s = pxm.data.dummy_data.get_simple_beam_shift_signal()
        >>> s_rot = s.rotate_scan_dimensions(10)
        >>> s_rot.plot()

        """
        s_temp = self.T
        s_temp2 = s_temp.map(
            rotate,
            show_progressbar=False,
            inplace=False,
            reshape=reshape,
            angle=-angle,
        )
        s_new = s_temp2.T
        return s_new


class LazyBeamShift(BeamShift, LazySignal):
    _signal_type = "beam_shift"

    pass



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

from pyxem.generators import get_DisplacementGradientMap
from pyxem.signals.tensor_field import DisplacementGradientMap, StrainMap
import hyperspy.api as hs
import pytest
import numpy as np


def vector_operation(z, M):
    """
    Extracts our input format and applies operations vector by vector

    Parameters
    ----------
    z : (n,2) np.array
        input vectors
    M : (2,2) np.array
        Tranformation matrix,
    Returns
    -------
    z_transformed : (n,2) np.array
        Output vectors
    """
    v_transformed = []
    for i in np.arange(0, z.shape[0]):
        v_transformed.append(np.matmul(M, z[i, :]))
    return np.asarray(v_transformed)


def rotation(z):
    theta = np.deg2rad(3)
    c, s = np.cos(theta), np.sin(theta)
    R = np.asarray(([c, -s], [s, c]))
    return vector_operation(z, R)


def uniform_expansion(z):
    return 1.1 * z


def stretch_in_x(z):
    M = np.asarray([[1.1, 0], [0, 1]])
    return vector_operation(z, M)


def generate_test_vectors(v):
    """
    We imagine we measured 4 sets of vectors, from 4 regions of sample, normal,
    rotated, uniform expansion and uniaxial expansion.
    """
    return np.asarray([[v, rotation(v)], [uniform_expansion(v), stretch_in_x(v)]])


def generate_displacment_map(vectors, return_residuals=False):
    deformed = hs.signals.Signal2D(generate_test_vectors(vectors))
    return get_DisplacementGradientMap(
        deformed, vectors, return_residuals=return_residuals
    )


def generate_strain_map(vectors):
    return generate_displacment_map(vectors).get_strain_maps()


class TestDisplacementGradientMap:
    four_vectors = np.asarray([[1, 0], [0, 1], [1, -1], [1, 1]])
    left_handed = np.asarray(([1, 1], [1, -1.2]))
    xy_vectors = np.asarray([[1, 0], [0, 1]])

    def setup_method(self):
        # reinitalizing to use inside functions...
        self.left_handed = np.asarray(([1, 1], [1, -1.2]))
        self.xy_vectors = np.asarray([[1, 0], [0, 1]])
        self.four_vectors = np.asarray([[1, 0], [0, 1], [1, -1], [1, 1]])

    @pytest.mark.parametrize("four_vectors", (four_vectors,))
    def test_generate_test_vectors(self, four_vectors):
        test_vectors = generate_test_vectors(four_vectors)
        assert np.shape(test_vectors) == (2, 2, 4, 2)
        assert np.shape(test_vectors[0, 0]) == four_vectors.shape
        assert np.shape(test_vectors[0, 1]) == four_vectors.shape

    @pytest.mark.parametrize("residuals", (True, False))
    @pytest.mark.parametrize("vectors", (four_vectors, xy_vectors, left_handed))
    def test_generate_displacement_map(self, vectors, residuals):
        if residuals:
            dis_map, r = generate_displacment_map(vectors, return_residuals=residuals)
        else:
            dis_map = generate_displacment_map(vectors, return_residuals=residuals)
        assert isinstance(dis_map, DisplacementGradientMap)

    @pytest.mark.parametrize("vectors", (four_vectors, xy_vectors, left_handed))
    def test_generate_strain_map(self, vectors):
        dis_map = generate_strain_map(vectors)
        assert isinstance(dis_map, StrainMap)

    def test_results_returned_correctly_in_same_basis(self):
        """Basic test of the summary statement for this section"""
        xy = generate_strain_map(self.xy_vectors)
        lh = generate_strain_map(self.left_handed)
        fv = generate_strain_map(self.four_vectors)
        np.testing.assert_almost_equal(xy.data, lh.data, decimal=2)
        np.testing.assert_almost_equal(xy.data, fv.data, decimal=2)

    def test_trivial_weight_function_case(self):
        """If weights are [1,1,1,1] the result should be the same as weights=None"""
        weights = [1, 1, 1, 1]
        xy = generate_strain_map(self.xy_vectors)
        deformed = hs.signals.Signal2D(generate_test_vectors(self.four_vectors))
        weight_strain_map = get_DisplacementGradientMap(
            deformed, self.four_vectors, weights=weights
        ).get_strain_maps()
        np.testing.assert_almost_equal(xy.data, weight_strain_map.data, decimal=2)

    def test_weight_function_behaviour(self):
        """Confirms that  a weight function [1,1,2,2] on [a,a,b,b] gives (2a+4b)/6 as the strain"""
        multi_vector_array = self.four_vectors
        strained_by_1pc_in_x = vector_operation(
            multi_vector_array, np.asarray([[1.01, 0], [0, 1]])
        )  # first  2
        strained_by_2pc_in_x = vector_operation(
            multi_vector_array, np.asarray([[1.02, 0], [0, 1]])
        )  # second 2
        weights = [1, 1, 2, 2]  # ((0.1*2 + 0.2*4)/6) = 0.166666
        vectors = np.concatenate(
            (strained_by_1pc_in_x[:, :2], strained_by_2pc_in_x[:, 2:]), axis=1
        )
        deformed = hs.signals.Signal2D(
            np.asarray([[vectors, vectors], [vectors, vectors]])
        )
        strain_map = get_DisplacementGradientMap(
            deformed, multi_vector_array, weights=weights
        ).get_strain_maps()
        np.testing.assert_almost_equal(
            strain_map.inav[0].isig[0, 0].data[0], -1.0166666 + 1, decimal=2
        )


""" Each of these basis should return the same results, in an xy basis"""


class TestLazyNotLazy:
    def setup_method(self):
        data = np.empty((4, 2), dtype=object)
        for iy, ix in np.ndindex(data.shape):
            data[iy, ix] = np.array([[5, 10], [10, 5]])
        self.vs = hs.signals.BaseSignal(data, ragged=True)

    def test_not_lazy(self):
        vs = self.vs
        vs_ref = vs.data[0, 0]
        D = get_DisplacementGradientMap(vs, vs_ref)
        assert D.axes_manager.navigation_shape == (2, 4)
        assert D.axes_manager.signal_shape == (3, 3)
        D.get_strain_maps()

    def test_lazy(self):
        vs = self.vs.as_lazy()
        vs_ref = vs.data[0, 0]
        D = get_DisplacementGradientMap(vs, vs_ref)
        assert D.axes_manager.navigation_shape == (2, 4)
        assert D.axes_manager.signal_shape == (3, 3)
        D.get_strain_maps()



name = "pyxem"
version = "0.21.0"
author = "Duncan Johnstone, Phillip Crout, Carter Francis, Magnus Nord"
copyright = "Copyright 2016-2025, The pyxem developers"
# Contributors listed by original committer, maintainers, then other
# within groups organised by  number of commits.
# See: https://github.com/pyxem/pyxem/graphs/contributors
credits = [
    "Duncan Johnstone",
    "Phillip Crout",
    "Carter Francis",
    "Magnus Nord",
    "Joonatan Laulainen",
    "Simon Høgås",
    "Eirik Opheim",
    "Eric Prestat",
    "Ben Martineau",
    "Tina Bergh",
    "Viljar Johan Femoen",
    "Niels Cautaerts",
    "Håkon Wiik Ånes",
    "Stef Smeets",
    "Andrew Ross",
    "Joseph Broussard",
    "Sivert Dagenborg",
    "Shuoyuan Huang",
    "Sean Collins",
    "Tom Furnival",
    "Daen Jannis",
    "Ida Hjorth",
    "Endre Jacobsen",
    "Mohsen Danaie",
    "Andrew Herzing",
    "Tim Poon",
    "Ruben Bjørge",
    "Tiarnan Doherty",
    "Affan Iqbal",
    "Jedrzej Morzy",
    "Tomas Ostasevicius",
    "Tor Inge Thorsen",
    "Matt von Lany",
    "Rob Tovey",
    "Petr Vacek",
]
license = "GPLv3+"
maintainer = "Duncan Johnstone, Phillip Crout, Magnus Nord, Carter Francis"
email = "pyxem.team@gmail.com"
status = "development"



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Tools for ellipse fitting using RANSAC."""

from tqdm import tqdm
import math
from functools import partial
import numpy as np
from skimage.measure import EllipseModel, ransac
import warnings
from hyperspy.signals import BaseSignal
from hyperspy.misc.utils import isiterable
import hyperspy.api as hs

__all__ = [
    "is_ellipse_good",
    "make_ellipse_data_points",
    "ellipse_to_markers",
    "get_ellipse_model_ransac",
    "get_ellipse_model_ransac_single_frame",
    "determine_ellipse",
]


def is_ellipse_good(
    ellipse_model,
    data,
    xf,
    yf,
    rf_lim,
    semi_len_min=None,
    semi_len_max=None,
    semi_len_ratio_lim=None,
):
    """Check if an ellipse model is within parameters.

    Parameters
    ----------
    ellipse_model : skimage EllipseModel
    data : Not used
    xf, yf : scalar
        Focus of the ellipse
    rf_lim : scalar
        If the distance from (xf, yf) and the centre of the
        ellipse is larger than rf_lim, False is returned.
    semi_len_min, semi_len_max : scalar
        Minimum and maximum semi length values for the ellipse, if any
        of the two semi lengths are outside this range, False is returned.
    semi_len_ratio_lim : scalar
        If the ratio between the largest and smallest semi length is larger
        than semi_len_ratio_lim, False is returned

    Returns
    -------
    is_good : bool

    Examples
    --------
    >>> import pyxem.utils.ransac_ellipse_tools as ret
    >>> model = ret.EllipseModel()
    >>> model.params = ret._make_ellipse_model_params_focus(30, 50, 30, 20, 0)
    >>> is_good = ret.is_ellipse_good(
    ...         ellipse_model=model, data=None, xf=30, yf=50, rf_lim=5)

    """
    xc, yc, a, b, r = ellipse_model.params
    x, y = _get_closest_focus(xf, yf, xc, yc, a, b, r)
    rf = math.hypot(x - xf, y - yf)
    if rf > rf_lim:
        return False
    if semi_len_min is not None:
        if a < semi_len_min:
            return False
        if b < semi_len_min:
            return False
    if semi_len_max is not None:
        if a > semi_len_max:
            return False
        if b > semi_len_max:
            return False
    if semi_len_ratio_lim is not None:
        semi_len_ratio = max(a, b) / min(a, b)
        if semi_len_ratio > semi_len_ratio_lim:
            return False
    return True


def _ellipse_centre_to_focus(x, y, a, b, r):
    """Get focus for an ellipse from EllipseModel.params

    Parameters
    ----------
    x, y : scalars
        Centre position of the ellipse.
    a, b : scalars
        Semi lengths
    r : scalar
        Rotation, in theta

    Returns
    -------
    foci : tuple of tuples
        (x focus 0, y focus 0), (x focus 1, y focus 1)

    Examples
    --------
    >>> import pyxem.utils.ransac_ellipse_tools as ret
    >>> f0, f1 = ret._ellipse_centre_to_focus(20, 32, 12, 9, 0.2)

    """
    if a < b:
        r += math.pi / 2
        a, b = b, a
    c = math.sqrt(math.pow(a, 2) - math.pow(b, 2))
    xf0, yf0 = x + c * math.cos(r), y + c * math.sin(r)
    xf1, yf1 = x - c * math.cos(r), y - c * math.sin(r)
    return ((xf0, yf0), (xf1, yf1))


def _get_closest_focus(x, y, xc, yc, a, b, r):
    """Get the focus closest to the centre from EllipseModel parameters

    Parameters
    ----------
    x, y : scalar
        Centre position of the diffraction pattern.
    xc, yc : scalars
        Centre position of the ellipse.
    a, b : scalars
        Semi lengths
    r : scalar
        Rotation, in theta

    Returns
    -------
    xf, yf : tuple
        Ellipse focus closest to the diffraction centre

    Examples
    --------
    >>> import pyxem.utils.ransac_ellipse_tools as ret
    >>> xf, yf = ret._get_closest_focus(25, 30, 20, 32, 12, 9, 0.2)

    """
    (xf0, yf0), (xf1, yf1) = _ellipse_centre_to_focus(xc, yc, a, b, r)
    rf0 = math.hypot(x - xf0, y - yf0)
    rf1 = math.hypot(x - xf1, y - yf1)
    if rf0 <= rf1:
        xf, yf = xf0, yf0
    else:
        xf, yf = xf1, yf1
    return (xf, yf)


def make_ellipse_data_points(x, y, a, b, r, nt=20, use_focus=True):
    """Get an ellipse position list.

    Parameters
    ----------
    x, y : scalars
        Centre position of the ellipse.
    a, b : scalars
        Semi lengths
    r : scalar
        Rotation, in theta
    nt : int, optional
        Number of data positions, default 20
    use_focus : bool
        If True, (x, y) will be the focus. If False,
        (x, y) will be centre of the ellipse.

    Returns
    -------
    data : numpy.ndarray
        [[x0, y0], [x1, y1], ...]

    Examples
    --------
    >>> import pyxem.utils.ransac_ellipse_tools as ret
    >>> data = ret.make_ellipse_data_points(5, 9, 8, 4, np.pi/3)

    Using all the arguments

    >>> data = ret.make_ellipse_data_points(5, 9, 8, 4, 0, nt=40,
    ...                                     use_focus=False)

    """
    if use_focus:
        params = _make_ellipse_model_params_focus(x, y, a, b, r)
    else:
        params = (x, y, a, b, r)
    theta_array = np.arange(0, 2 * np.pi, 2 * np.pi / nt)
    data = EllipseModel().predict_xy(theta_array, params=params)
    return data


def _ellipse_model_centre_to_focus(xc, yc, a, b, r, x, y):
    """Get focus params from centre params

    There are two different focuses, the one closest to
    (x, y) is returned.

    Parameters
    ----------
    xc, yc : scalar
        x and y position of the ellipse centre
    a, b : scalar
        Semi lengths
    r : scalar
        Rotation
    x, y : scalar

    Returns
    -------
    params : tuple
        (xf, yf, a, b, r)

    """
    xf, yf = _get_closest_focus(x, y, xc, yc, a, b, r)
    params = (xf, yf, a, b, r)
    return params


def _make_ellipse_model_params_focus(xf, yf, a, b, r):
    """
    Parameters
    ----------
    xf, yf : scalar
        x and y position of the focus
    a, b : scalar
        Semi lengths
    r : scalar
        Rotation

    Returns
    -------
    params : tuple
        (xc, yc, a, b, r)

    """
    if a < b:
        r += math.pi / 2
        a, b = b, a
    c = math.sqrt(math.pow(a, 2) - math.pow(b, 2))
    xc = xf - c * math.cos(r)
    yc = yf - c * math.sin(r)
    params = (xc, yc, a, b, r)
    return params


def get_ellipse_model_ransac_single_frame(
    data,
    xf=128,
    yf=128,
    rf_lim=30,
    semi_len_min=50,
    semi_len_max=90,
    semi_len_ratio_lim=1.2,
    min_samples=6,
    residual_threshold=10,
    max_trials=500,
):
    """Pick a random number of data points to fit an ellipse to.

    The ellipse's constraints can be specified.

    See skimage.measure.ransac for more information.

    Parameters
    ----------
    data : numpy.ndarray
        In the form [[x0, y0], [x1, y1], ...]
    xf, yf : scalar, optional
        Default 128
    rf_lim : scalar, optional
        How far the ellipse centre can be from (xf, yf)
    semi_len_min, semi_len_max : scalar, optional
        Limits of the semi lengths
    semi_len_ratio_lim : scalar, optional
        Limit of the ratio of the semi length, must be equal or larger
        than 1. This ratio is calculated by taking the largest semi length
        divided by the smallest semi length:
        max(semi0, semi1)/min(semi0, semi1). So for a perfect circle this
        ratio will be 1.
    min_samples : scalar, optional
        Minimum number of data points to fit the ellipse model to.
    residual_threshold : scalar, optional
        Maximum distance for a data point to be considered an inlier.
    max_trials : scalar, optional
        Maximum number of tries for the ransac algorithm.

    Returns
    -------
    model_ransac, inliers
        Model data is accessed in model_ransac.params:
        [x, y, semi_len0, semi_len1, rotation]

    Examples
    --------
    >>> import pyxem.utils.ransac_ellipse_tools as ret
    >>> data = ret.EllipseModel().predict_xy(
    ...        np.arange(0, 2*np.pi, 0.5), params=(128, 130, 50, 60, 0.2))
    >>> ellipse_model, inliers = ret.get_ellipse_model_ransac_single_frame(
    ...        data, xf=128, yf=128, rf_lim=5, semi_len_min=45,
    ...        semi_len_max=65, semi_len_ratio_lim=1.4, max_trials=1000)

    """
    is_model_valid = partial(
        is_ellipse_good,
        xf=xf,
        yf=yf,
        rf_lim=rf_lim,
        semi_len_min=semi_len_min,
        semi_len_max=semi_len_max,
        semi_len_ratio_lim=semi_len_ratio_lim,
    )
    if min_samples > len(data):
        min_samples = len(data) - 1
    # This for loop is here to avoid the returned model being outside the
    # specified limits especially semi_len_ratio_lim.
    # This can happen if only the ransac function is used with this check.
    # This is (probably) due to a valid model being found first, then
    # additional inliers are found afterwards.
    for i in range(3):
        model_ransac, inliers = ransac(
            data.astype(np.float32),
            EllipseModel,
            min_samples=min_samples,
            residual_threshold=residual_threshold,
            max_trials=max_trials,
            is_model_valid=is_model_valid,
        )
        if model_ransac is not None:
            if is_model_valid(model_ransac, None):
                break
            else:
                model_ransac, inliers = None, None
        else:
            break
    return model_ransac, inliers


def get_ellipse_model_ransac(
    data,
    xf=128,
    yf=128,
    rf_lim=30,
    semi_len_min=70,
    semi_len_max=90,
    semi_len_ratio_lim=1.2,
    min_samples=6,
    residual_threshold=10,
    max_trials=500,
    show_progressbar=True,
):
    """Pick a random number of data points to fit an ellipse to.

    The ellipse's constraints can be specified.

    See skimage.measure.ransac for more information.

    Parameters
    ----------
    data : numpy.ndarray
        In the form [[[[x0, y0], [x1, y1], ...]]]
    xf, yf : scalar, optional
        Default 128 center of the diffraction pattern
    rf_lim : scalar, optional
        How far the ellipse centre can be from (xf, yf)
    semi_len_min, semi_len_max : scalar, optional
        Limits of the semi lengths
    semi_len_ratio_lim : scalar, optional
        Limit of the ratio of the semi length, must be equal or larger
        than 1. This ratio is calculated by taking the largest semi length
        divided by the smallest semi length:
        max(semi0, semi1)/min(semi0, semi1). So for a perfect circle this
        ratio will be 1.
    min_samples : scalar, optional
        Minimum number of data points to fit the ellipse model to.
    residual_threshold : scalar, optional
        Maximum distance for a data point to be considered an inlier.
    max_trials : scalar, optional
        Maximum number of tries for the ransac algorithm.
    show_progressbar : bool, optional
        Default True

    Returns
    -------
    ellipse_array, inlier_array : numpy.ndarray
        Model data is accessed in ellipse_array, where each probe position
        (for two axes) contain a list with the ellipse parameters:
        [y, x, semi_len0, semi_len1, rotation]. If no ellipse is found
        this is None.

    """
    if not isiterable(xf):
        xf = np.ones(data.shape[:2]) * xf
    if not isiterable(yf):
        yf = np.ones(data.shape[:2]) * yf

    ellipse_array = np.zeros(data.shape[:2], dtype=object)
    inlier_array = np.zeros(data.shape[:2], dtype=object)
    num_total = data.shape[0] * data.shape[1]
    t = tqdm(np.ndindex(data.shape[:2]), disable=not show_progressbar, total=num_total)

    new_peaks = np.empty(data.shape, dtype=object)
    for i in np.ndindex(data.shape):
        new_peaks[i] = data[i][:, ::-1]

    for iy, ix in t:
        temp_xf, temp_yf = xf[iy, ix], yf[iy, ix]
        ellipse_model, inliers = get_ellipse_model_ransac_single_frame(
            new_peaks[iy, ix],  # reverse x,y for pixel units
            xf=temp_xf,
            yf=temp_yf,
            rf_lim=rf_lim,
            semi_len_min=semi_len_min,
            semi_len_max=semi_len_max,
            semi_len_ratio_lim=semi_len_ratio_lim,
            min_samples=min_samples,
            residual_threshold=residual_threshold,
            max_trials=max_trials,
        )
        if ellipse_model is not None:
            params = ellipse_model.params
        else:
            params = None
        ellipse_array[iy, ix] = params
        inlier_array[iy, ix] = inliers
    return ellipse_array, inlier_array


def _get_max_positions(signal, mask=None, num_points=5000):
    """Gets the top num_points pixels in the dataset.

    Parameters
    --------------
    signal : BaseSignal
        The signal which we want to find the max positions for.
    mask : np.array
        A mask to be applied to the data for values to ignore.
    num_points : int
        The number of points to be considered.

    """
    if isinstance(signal, BaseSignal):
        data = signal.data
    else:
        data = signal
    i_shape = np.shape(data)
    flattened_array = data.flatten()
    if mask is not None:
        flattened_mask = mask.flatten()
        flattened_array[flattened_mask] = 0
    # take top 5000 points make sure exclude zero beam
    indexes = np.argsort(flattened_array)
    cords = np.array(
        [
            np.floor_divide(indexes[-num_points:], i_shape[1]),
            np.remainder(indexes[-num_points:], i_shape[1]),
        ]
    )  # [x axis (row),y axis (col)]
    return cords.T


def _ellipse_to_affine(major, minor, rot):
    if minor > major:
        temp = minor
        minor = major
        major = temp
        rot = rot + np.pi / 2
    rot = np.pi / 2 - rot
    if rot < 0:
        rot = rot + np.pi

    Q = [[np.cos(rot), -np.sin(rot), 0], [np.sin(rot), np.cos(rot), 0], [0, 0, 1]]
    S = [[1, 0, 0], [0, minor / major, 0], [0, 0, 1]]
    C = np.matmul(np.matmul(Q, S), np.transpose(Q))
    return C


def mask_peak_array(array, mask, invert=False):
    """Return only the peaks in the array which are not masked. Works for
    both ragged and non-ragged arrays.

    Parameters
    ----------
    array: np.ndarray
        The array of peaks to be masked. Can be ragged or non-ragged.
    mask: np.ndarray
        The mask to be applied to the array. If the array is ragged, the mask
        must be ragged as well.
    invert: bool
        If True, the mask is inverted.
    """
    if array.dtype != object:
        if invert:
            mask = np.logical_not(mask)
        return array[mask]
    else:
        masked_array = np.empty(array.shape, dtype=object)
        for i in np.ndindex(array.shape):
            if invert:
                m = np.logical_not(mask[i])
            else:
                m = mask[i]
            masked_array[i] = array[i][m]
        return masked_array


def ellipse_to_markers(ellipse_array, points=None, inlier=None):
    """Convert an ellipse array to a :class:`hyperspy.api.plot.markers.Ellipses` object. If points and
    inlier are provided, then the points are also plotted. The inlier points are plotted in green
    and the outlier points are plotted in red.

    Parameters
    ----------
    ellipse_array: np.ndarray
        The array of ellipses parameters in the form [x_c, y_c, semi_len0, semi_len1, rotation]
    points: np.ndarray
        The array of points to be plotted. If None, then no points are plotted.
    inlier:np.ndarray
        The bool array of inlier points. If None, then no points are plotted.

    Returns
    -------

    """
    if not isinstance(ellipse_array, np.ndarray):
        ellipse_array = np.array(ellipse_array)
    ellipse_array = ellipse_array.T
    if ellipse_array.dtype == object:
        offsets = np.empty(ellipse_array.shape, dtype=object)
        heights = np.empty(ellipse_array.shape, dtype=object)
        widths = np.empty(ellipse_array.shape, dtype=object)
        angles = np.empty(ellipse_array.shape, dtype=object)
        for i in np.ndindex(ellipse_array.shape):
            offsets[i] = ellipse_array[i][:2][::-1]
            heights[i] = ellipse_array[i][2] * 2
            widths[i] = ellipse_array[i][3] * 2
            angles[i] = np.rad2deg(-ellipse_array[i][4])
    else:
        offsets = np.array(
            [
                ellipse_array[:2][::-1],
            ]
        )
        heights = ellipse_array[2] * 2
        widths = ellipse_array[3] * 2
        angles = np.rad2deg(-ellipse_array[4])

    el = hs.plot.markers.Ellipses(
        offsets=offsets,
        heights=heights,
        widths=widths,
        angles=angles,
        facecolor="none",
        edgecolor="white",
        lw=4,
    )

    if points is not None and inlier is not None:
        in_points = hs.plot.markers.Points(
            offsets=mask_peak_array(points[:, ::-1], inlier), color="green"
        )
        out_points = hs.plot.markers.Points(
            offsets=mask_peak_array(points[:, ::-1], inlier, invert=True),
            color="red",
            alpha=0.5,
        )
        return el, in_points, out_points
    elif points is not None:
        points = hs.plot.markers.Points(
            offsets=points[:, ::-1], color="green", alpha=0.5
        )
        return el, points
    else:
        return el


def determine_ellipse(
    signal=None,
    pos=None,
    mask=None,
    num_points=1000,
    use_ransac=False,
    guess_starting_params=True,
    return_params=False,
    **kwargs,
):
    """
    This method starts by taking some number of points which are the most intense
    in the signal or those points can be directly passed.  It then takes those points
    and guesses some starting parameters for the `get_ellipse_model_ransac_single_frame`
    function. From there it will try to determine the ellipse parameters.

    Parameters
    -----------
    signal : Signal2D
        The signal of interest.
    pos : np.ndarray
        The positions of the points to be used to determine the ellipse.
    mask : Array-like
        The mask to be applied to the data.  The True values are ignored.
    num_points : int
        The number of points to consider.
    use_ransac : bool
        If Ransac should be used to determine the ellipse. False is faster but less.
        robust with respect to noise.
    guess_starting_params : bool
        If True then the starting parameters will be guessed based on the points determined.
    return_params : bool
        If the ellipse parameters should be returned as well.
    **kwargs:
        Any other keywords for ``get_ellipse_model_ransac_single_frame``.

    Returns
    -------
    center : (x,y)
        The center of the diffraction pattern.
    affine :
        The affine transformation to make the diffraction pattern circular.

    Examples
    --------
    >>> import pyxem.utils.ransac_ellipse_tools as ret
    >>> import pyxem.data.dummy_data.make_diffraction_test_data as mdtd
    >>> test_data = mdtd.MakeTestData(200, 200, default=False)
    >>> test_data.add_disk(x0=100, y0=100, r=5, intensity=30)
    >>> test_data.add_ring_ellipse(x0=100, y0=100, semi_len0=63, semi_len1=70, rotation=45)
    >>> s = test_data.signal
    >>> s.set_signal_type("electron_diffraction")
    >>> import numpy as np
    >>> mask = np.zeros_like(s.data, dtype=bool)
    >>> mask[100 - 20:100 + 20, 100 - 20:100 + 20] = True # mask beamstop
    >>> center, affine = ret.determine_ellipse(s, mask=mask, use_ransac=False)
    >>> s_corr = s.apply_affine_transformation(affine, inplace=False)

    """
    if signal is not None:
        pos = _get_max_positions(signal, mask=mask, num_points=num_points)
    elif pos is None:
        raise ValueError("Either signal or pos must be specified")
    if use_ransac:
        if guess_starting_params:
            el, inlier = get_ellipse_model_ransac_single_frame(
                pos,
                xf=np.mean(pos[:, 0]),
                yf=np.mean(pos[:, 1]),
                rf_lim=np.shape(signal.data)[0] / 5,
                semi_len_min=np.std(pos[:, 1]),
                semi_len_max=np.std(pos[:, 1]) * 2,
                semi_len_ratio_lim=1.2,
                min_samples=6,
                residual_threshold=20,
                max_trials=1000,
            )
        else:
            el, inlier = get_ellipse_model_ransac_single_frame(pos, **kwargs)
    else:
        e = EllipseModel()
        converge = e.estimate(data=pos)
        el = e
    if el is not None:
        affine = _ellipse_to_affine(el.params[3], el.params[2], el.params[4])
        center = (el.params[0], el.params[1])
        if return_params and use_ransac:
            return center, affine, el.params, pos, inlier
        elif return_params:
            return center, affine, el.params, pos
        else:
            return center, affine
    else:  # pragma: no cover
        warnings.warn("Ransac Ellipse detection did not converge")
        return None



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from hyperspy.component import Component
from diffsims.utils.lobato_scattering_params import ATOMIC_SCATTERING_PARAMS_LOBATO


class ScatteringFitComponentLobato(Component):
    """
    A scattering component to fit background atomic scattering.
    Calculates the sum of the squares sum_squares = sum (ci * fi**2 )
    and the square of the sum square_sum = (sum(ci * fi))**2
    for atomic fraction ci with electron scattering factor fi.
    The latter is used for normalisation.

    N and C are fitting parameters for the Component class.

    The fitted function f is of the form:

    f = sum(fi), fi = ai(2 + bi*g^2) / (1 + bi*g^2)^2
    where 1 < i < 5, ai, and bi are tabulated constants, and
    g = 1/d = 2sin(theta)/lambda is the scattering vector magnitude.
    The factors are tabulated in ATOMIC_SCATTERING_PARAMS_LOBATO.


    The parametrisation is detailed in [1].

    Parameters
    ----------
    elements: list of str
        A list of elements present (by symbol).
    fracs: list of float
        A list of fraction of the respective elements. Should sum to 1.
    N : float
        The "slope" of the fit.
    C : float
        An additive constant to the fit.

    References
    ----------
    [1] Lobato, I., & Van Dyck, D. (2014). An accurate parameterization for
    scattering factors, electron densities and electrostatic potentials for
    neutral atoms that obey all physical constraints. Acta Crystallographica
    Section A: Foundations and Advances, 70(6), 636-649.

    """

    def __init__(self, elements, fracs, N=1.0, C=0.0):
        Component.__init__(self, ["N", "C"])
        self._whitelist["elements"] = ("init,sig", elements)
        self._whitelist["fracs"] = ("init,sig", fracs)
        self.elements = elements
        self.fracs = fracs
        params = []
        for e in elements:
            params.append(ATOMIC_SCATTERING_PARAMS_LOBATO[e])
        self.params = params
        self.N.value = N
        self.C.value = C

    def function(self, x):
        """
        The actual fitted function. Based on the HyperSpy Component class.
        The function "function" is called innately by HyperSpy's fitting
        algorithm, and represents the appropriate atomic scattering factor
        function presented respectively in Lobato & Van Dyck (2014).

        Parameters
        ----------
        x : array-like
            The scattering vector magnitude

        """
        N = self.N.value
        C = self.C.value
        params = self.params
        fracs = self.fracs

        sum_squares = np.zeros(x.size)
        square_sum = np.zeros(x.size)

        for i, element in enumerate(params):
            fi = np.zeros(x.size)
            for n in range(len(element)):  # 5 parameters per element
                fi += (
                    element[n][0]
                    * (2 + element[n][1] * np.square(x))
                    * np.divide(1, np.square(1 + element[n][1] * np.square(x)))
                )
            elem_frac = fracs[i]
            sum_squares += np.square(fi) * elem_frac
            square_sum += fi * elem_frac

        self.square_sum = np.square(square_sum)
        # square sum is kept for normalisation.
        return N * sum_squares + C
